# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import math

import numpy as np


class Box(object):
    def __init__(
        self,
        width=1,
        height=1,
        length=1,
        centerX=0,
        centerY=0,
        centerZ=0,
        yaw=0.0,
        pitch=0.0,
        roll=0.0,
        translationX=0,
        translationY=0,
        translationZ=0,
    ):
        # In webots length is in z-axis, width is in x-axis and height is in y-axis
        # Center is the rotation center for the box
        #  -> in webots, this should be the rear axle location relative to the center of the box
        #  -> center is the vector from the true center of the box to the rotation center of the box
        # In webots yaw is CC around the y-axis!
        # In webots pitch is CC around the z-axis!
        # In webots roll is CC around the x-axis!
        # NOTE: this geometry class applies a translation to get the center of rotation,
        #  rotates the box and then applies a global translation to move the rectangle in a global coordinate system
        self.dimensions = np.array([width, height, length])
        self.center = np.array([centerX, centerY, centerZ])
        self.translation = np.array([translationX, translationY, translationZ])
        self.yaw = yaw
        self.pitch = pitch
        self.roll = roll
        self.unrotatedegocorners = self._getunrotatedegocorners()
        self.rotation = self.getyawrollpitchrotation(self.yaw, self.pitch, self.roll)
        # The transpose is the inverse rotation matrix
        self.reverserotation = np.transpose(self.rotation)
        self.corners = self.getcorners()

    def __str__(self):
        return "[({},{},{}), center=({},{},{}), rotation=({},{},{}), translation=({},{},{})]".format(
            self.dimensions[0],
            self.dimensions[1],
            self.dimensions[2],
            self.center[0],
            self.center[1],
            self.center[2],
            self.yaw,
            self.pitch,
            self.roll,
            self.translation[0],
            self.translation[1],
            self.translation[2],
        )

    def getyawrollpitchrotation(self, yaw, pitch, roll):
        sin_p = math.sin(pitch)
        cos_p = math.cos(pitch)
        sin_y = math.sin(yaw)
        cos_y = math.cos(yaw)
        sin_r = math.sin(roll)
        cos_r = math.cos(roll)
        return np.array(
            [
                [
                    cos_p * cos_y,
                    cos_p * sin_y * sin_r - sin_p * cos_r,
                    cos_p * sin_y * cos_r + sin_p * sin_r,
                ],
                [
                    sin_p * cos_y,
                    sin_p * sin_y * sin_r + cos_p * cos_r,
                    sin_p * sin_y * cos_r - cos_p * sin_r,
                ],
                [-sin_y, cos_y * sin_r, cos_y * cos_r],
            ]
        )

    def _getunrotatedegocorners(self):
        x_diff1, y_diff1, z_diff1 = -self.dimensions / 2.0 - self.center
        x_diff2, y_diff2, z_diff2 = self.dimensions / 2.0 - self.center
        x1, y1, z1 = [
            min(x_diff1, x_diff2),
            min(y_diff1, y_diff2),
            min(z_diff1, z_diff2),
        ]
        x2, y2, z2 = [
            max(x_diff1, x_diff2),
            max(y_diff1, y_diff2),
            max(z_diff1, z_diff2),
        ]
        corners = np.array(
            [
                [x1, y1, z1],
                [x1, y1, z2],
                [x1, y2, z1],
                [x1, y2, z2],
                [x2, y1, z1],
                [x2, y1, z2],
                [x2, y2, z1],
                [x2, y2, z2],
            ]
        )
        return corners

    def getcorners(self):
        corners = self._getunrotatedegocorners()

        if abs(self.yaw) > 1e-30 or abs(self.pitch) > 1e-30 or abs(self.roll) > 1e-30:
            corners = np.inner(corners, self.rotation)

        corners += self.translation

        return corners

    def getvolume(self):
        return np.prod(self.dimensions)

    def containspoint(self, point):
        return self.containspoints(np.array([point]))

    def containspoints(self, points):
        # 1.) Rotate the point around the center
        # 2.) Check to see if the points lie inside the co-linear rectangle
        N, d = points.shape
        ego_points = points - self.translation
        if abs(self.yaw) > 1e-30 or abs(self.pitch) > 1e-30 or abs(self.roll) > 1e-30:
            rotated_points = np.inner(ego_points, self.reverserotation)
        else:
            rotated_points = ego_points
        low_corner = self.unrotatedegocorners[0]
        high_corner = self.unrotatedegocorners[7]

        # This is why we rotate the points rather than the box -> simpler to check if the box is
        #  co-linear with the axis of the local coordinate system
        return np.all(
            np.logical_and(
                (high_corner >= rotated_points), (rotated_points >= low_corner)
            ),
            axis=1,
        )

    # Note to be used externly
    def _unrotated_containspoints(self, unrotated_points):
        low_corner = self.unrotatedegocorners[0]
        high_corner = self.unrotatedegocorners[7]

        # This is why we rotate the points rather than the box -> simpler to check if the box is
        #  co-linear with the axis of the local coordinate system
        return np.all(
            np.logical_and(
                (high_corner >= unrotated_points), (unrotated_points >= low_corner)
            ),
            axis=1,
        )

    def _getnormals(self):
        # Just need three normals of the unrotated box
        p1, p2, p3, p4, p5, p6, p7, p8 = self.unrotatedegocorners
        xn = np.cross(p3 - p1, p2 - p1)
        yn = np.cross(p2 - p1, p5 - p1)
        zn = np.cross(p5 - p1, p3 - p1)

        return xn, yn, zn

    def getlines(self):
        p1, p2, p3, p4, p5, p6, p7, p8 = self.corners
        start_points = np.array([p1, p1, p1, p2, p2, p3, p3, p4, p5, p5, p6, p7])
        end_points = np.array([p2, p3, p5, p4, p6, p7, p4, p8, p6, p7, p8, p8])
        return start_points, end_points

    def intersects(self, box):
        # NOTE: the order of the points in self.corners and self.unrotatedegocorners must not change!

        # Calculates whether any corners of rect fall within self
        start1, end1 = box.getlines()
        intersect1 = self.intersectswithlines(points=start1, end_points=end1)

        # Also need to see if any of the corners of self fall in rect
        start2, end2 = self.getlines()
        intersect2 = box.intersectswithlines(points=start2, end_points=end2)

        return np.any(np.concatenate((intersect1, intersect2)))

    # Calculates intersection point between two parallel planes with norm and defined by points 1 and 2 respectively
    # norm must be the outer norm for plane1 defined by point pts_on_plane1
    def _get_line_intersect_with_planes_3d(
        self, points, directions, norm, pts_on_plane1, pts_on_plane2
    ):
        r = directions
        n1 = norm
        n2 = -norm
        d1 = -np.inner(n1, pts_on_plane1[0])
        d2 = -np.inner(n2, pts_on_plane2[0])

        r_n1 = np.inner(r, n1)
        r_n2 = np.inner(r, n2)
        n1_px = np.inner(n1, points) + d1
        n2_px = np.inner(n2, points) + d2
        n1_p = np.inner(n1, points)
        n2_p = np.inner(n2, points)

        t1 = np.zeros(len(points))
        t2 = np.zeros(len(points))

        # Check for parallel
        z1 = np.abs(r_n1) < 1e-20
        z2 = np.abs(r_n2) < 1e-20
        nz1 = np.logical_not(z1)
        nz2 = np.logical_not(z2)
        # Check for points on plane
        on1 = np.abs(n1_px) < 1e-20
        on2 = np.abs(n2_px) < 1e-20
        non1 = np.logical_not(on1)
        non2 = np.logical_not(on2)

        # All points that are not on the plane but are perpendicular -> inf
        t1[np.logical_and(z1, non1)] = -np.inf
        t2[np.logical_and(z2, non2)] = np.inf

        # All points not perpendicular and not on the plane
        nz_non1 = np.logical_and(nz1, non1)
        nz_non2 = np.logical_and(nz2, non2)
        t1[nz_non1] = -(d1 + n1_p[nz_non1]) / r_n1[nz_non1]
        t2[nz_non2] = -(d2 + n2_p[nz_non2]) / r_n2[nz_non2]

        # Re-order points if necessary
        t = np.stack((t1, t2), axis=1)
        tpos = np.min(t, axis=1)
        tneg = np.max(t, axis=1)

        # print("POS {}\nNEG {}".format(tpos, tneg))

        return tpos, tneg

    # NOTE: directions must be vectors from points (start) to end_points
    def intersectswithlines(self, points, end_points):
        # Method is described here: https://math.stackexchange.com/questions/1477930/does-a-line-intersect-a-box-in-a-3d-space
        rot_points = points - self.translation
        rot_end_points = end_points - self.translation
        if abs(self.yaw) > 1e-30 or abs(self.pitch) > 1e-30 or abs(self.roll) > 1e-30:
            rot_points = np.inner(rot_points, self.reverserotation)
            rot_end_points = np.inner(rot_end_points, self.reverserotation)

        rot_directions = rot_end_points - rot_points

        xn, yn, zn = self._getnormals()

        with np.errstate(divide="ignore"):
            low_xpoints = [self.unrotatedegocorners[0]]
            low_ypoints = [self.unrotatedegocorners[0]]
            low_zpoints = [self.unrotatedegocorners[0]]
            high_xpoints = [self.unrotatedegocorners[7]]
            high_ypoints = [self.unrotatedegocorners[7]]
            high_zpoints = [self.unrotatedegocorners[7]]

            t_xpos, t_xneg = self._get_line_intersect_with_planes_3d(
                rot_points, rot_directions, xn, high_xpoints, low_xpoints
            )
            t_ypos, t_yneg = self._get_line_intersect_with_planes_3d(
                rot_points, rot_directions, yn, high_ypoints, low_ypoints
            )
            t_zpos, t_zneg = self._get_line_intersect_with_planes_3d(
                rot_points, rot_directions, zn, high_zpoints, low_zpoints
            )

        pos_ts = np.stack((t_xpos, t_ypos, t_zpos), axis=1)
        neg_ts = np.stack((t_xneg, t_yneg, t_zneg), axis=1)

        # print("{} {}".format(pos_ts, neg_ts))

        maxpos = np.max(pos_ts, axis=1)
        minneg = np.min(neg_ts, axis=1)

        condition = np.array([False] * len(points))

        start_contains = self._unrotated_containspoints(rot_points)
        end_contains = self._unrotated_containspoints(rot_end_points)
        both = np.logical_and(start_contains, end_contains)
        one = np.logical_xor(start_contains, end_contains)
        none = np.logical_not(np.logical_or(both, one))

        # print("MAX {}; MIN {}".format(maxpos[none], minneg[none]))
        # print("POS {}; NEG {}".format(pos_ts[none], neg_ts[none]))

        # Handle the case where both points are in the box
        condition[both] = True
        # Handle the case where one point is in the box
        condition[one] = np.logical_and(
            maxpos[one] <= minneg[one],
            np.logical_and(maxpos[one] <= 1, minneg[one] >= 0),
        )
        # Handle the case where both points are outside the box
        if np.any(none):
            possibles = np.array([False] * len(points))
            possibles[none] = np.logical_and(
                maxpos[none] <= minneg[none],
                np.logical_and(
                    maxpos[none] >= 0,
                    np.logical_and(
                        minneg[none] <= 1,
                        np.logical_and(maxpos[none] <= 1, minneg[none] >= 0),
                    ),
                ),
            )
            if np.any(possibles):
                none_start_points = rot_points[possibles]
                none_directions = rot_directions[possibles]
                none_surface1 = none_start_points + np.transpose(
                    np.transpose(none_directions) * maxpos[possibles]
                )
                none_surface2 = none_start_points + np.transpose(
                    np.transpose(none_directions) * minneg[possibles]
                )
                # Update any possibles that were potentially true
                possibles[possibles] = np.logical_and(
                    self._unrotated_containspoints(none_surface1),
                    self._unrotated_containspoints(none_surface2),
                )
            condition[none] = possibles[none]

        return condition


# Support library for converting a LIDAR point (where the points are in scan order) to an image
class SphericalCartesianConverter(object):
    def __init__(self, hfov, vfov, width, height):
        # Assume tilt angle is 0!
        # Scan from left to right
        az = np.linspace(
            math.pi + hfov / 2.0, math.pi - hfov / 2.0, width, dtype=np.float32
        )
        # Scan from top layer to bottom layer
        el = np.linspace(vfov / 2.0, -vfov / 2.0, height, dtype=np.float32)
        az_grid, el_grid = np.meshgrid(az, el, sparse=True)
        self.z_factor = np.cos(az_grid) * np.cos(el_grid)
        self.x_factor = np.sin(az_grid) * np.cos(el_grid)
        self.y_factor = np.sin(el_grid)

    def calculate_point_cloud(self, depth):
        # These x, y, z are webots axes!
        z = depth * self.z_factor  # Points behind vehicle
        x = depth * self.x_factor  # Points to right of vehicle
        y = depth * self.y_factor  # Points up

        points = np.stack((x.flatten(), y.flatten(), z.flatten()), axis=-1)
        return points


def filter_points(points, ground_height=0.0, max_distance=99.9):
    distances = np.linalg.norm(points, axis=1)
    filtered_points = points[distances < max_distance]
    return filtered_points[filtered_points[:, 1] > ground_height]


def calculate_point_cloud(depth, hfov, vfov, width, height):
    # Assume tilt angle is 0!
    # Scan from left to right
    az = np.linspace(
        math.pi + hfov / 2.0, math.pi - hfov / 2.0, width, dtype=np.float32
    )
    # Scan from top layer to bottom layer
    el = np.linspace(vfov / 2.0, -vfov / 2.0, height, dtype=np.float32)
    az_grid, el_grid = np.meshgrid(az, el, sparse=True)
    z_factor = np.cos(az_grid) * np.cos(el_grid)
    x_factor = np.sin(az_grid) * np.cos(el_grid)
    y_factor = np.sin(el_grid)

    # These x, y, z are webots axes!
    z = depth * z_factor  # Points behind vehicle
    x = depth * x_factor  # Points to right of vehicle
    y = depth * y_factor  # Points up

    points = np.stack((x.flatten(), y.flatten(), z.flatten()), axis=-1)
    return points


def yaw_rotation(yaw):
    cos_y = math.cos(yaw)
    sin_y = math.sin(yaw)
    return np.array([[cos_y, 0, sin_y], [0, 1, 0], [-sin_y, 0, cos_y]])


def yaw_rotate(yaw, vectors):
    rotation = yaw_rotation(yaw)
    return np.inner(vectors, rotation)


# In webots yaw is CC around the y-axis!
# In webots pitch is CC around the z-axis!
# In webots roll is CC around the x-axis!
def yawrollpitch_rotation(yaw, pitch, roll):
    sin_p = math.sin(pitch)
    cos_p = math.cos(pitch)
    sin_y = math.sin(yaw)
    cos_y = math.cos(yaw)
    sin_r = math.sin(roll)
    cos_r = math.cos(roll)
    return np.array(
        [
            [
                cos_p * cos_y,
                cos_p * sin_y * sin_r - sin_p * cos_r,
                cos_p * sin_y * cos_r + sin_p * sin_r,
            ],
            [
                sin_p * cos_y,
                sin_p * sin_y * sin_r + cos_p * cos_r,
                sin_p * sin_y * cos_r - cos_p * sin_r,
            ],
            [-sin_y, cos_y * sin_r, cos_y * cos_r],
        ]
    )


def rotate(yaw, pitch, roll, vectors):
    rotation = yawrollpitch_rotation(yaw, pitch, roll)
    return np.inner(vectors, rotation)


def visualize_boxes(ax, boxes, markers=None, colors=None, faces_color=None):
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection

    if markers is None:
        markers = ["o"] * len(boxes)
    if colors is None:
        colors = ["red"]
    if faces_color is None:
        faces_color = ["cyan"]

    for box, marker, color, face_color in zip(boxes, markers, colors, faces_color):
        Z = box.corners
        ax.scatter(
            box.corners[:, 0],
            box.corners[:, 1],
            box.corners[:, 2],
            marker=marker,
            color=color,
        )

        # generate list of sides' polygons of our pyramid
        verts = [
            [Z[0], Z[1], Z[3], Z[2]],
            [Z[4], Z[5], Z[7], Z[6]],
            [Z[0], Z[1], Z[5], Z[4]],
            [Z[2], Z[3], Z[7], Z[6]],
            [Z[1], Z[5], Z[7], Z[3]],
            [Z[4], Z[0], Z[2], Z[6]],
        ]

        # plot sides
        ax.add_collection3d(
            Poly3DCollection(
                verts, facecolors=face_color, linewidths=0.5, edgecolors="k", alpha=0.5
            )
        )
    ax.set_zlim(0, 50)
    ax.set_xlim(50, 150)
    ax.set_ylim(50, 150)

    return ax



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import glob
import math
import os
from itertools import cycle
from sys import path

import numpy as np
import yaml, inspect
from scipy.spatial import distance

from smarts.core.scenario import Scenario
from smarts.env.hiway_env import HiWayEnv
from ultra.baselines.adapter import BaselineAdapter
from ultra.baselines.common.yaml_loader import load_yaml

path.append("./ultra")
from ultra.utils.common import ego_social_safety, get_closest_waypoint, get_path_to_goal


class UltraEnv(HiWayEnv):
    def __init__(
        self,
        agent_specs,
        scenario_info,
        headless,
        timestep_sec,
        seed,
        eval_mode=False,
        ordered_scenarios=False,
    ):
        self.timestep_sec = timestep_sec
        self.headless = headless
        self.scenario_info = scenario_info
        self.scenarios = self.get_task(scenario_info[0], scenario_info[1])
        if not eval_mode:
            _scenarios = glob.glob(f"{self.scenarios['train']}")
        else:
            _scenarios = glob.glob(f"{self.scenarios['test']}")

        self.ultra_scores = BaselineAdapter.reward_adapter

        super().__init__(
            scenarios=_scenarios,
            agent_specs=agent_specs,
            headless=headless,
            timestep_sec=timestep_sec,
            seed=seed,
            visdom=False,
        )

        if ordered_scenarios:
            scenario_roots = []
            for root in _scenarios:
                if Scenario.is_valid_scenario(root):
                    # The case that this is a scenario root
                    scenario_roots.append(root)
                else:
                    # The case that there this is a directory of scenarios: find each of the roots
                    scenario_roots.extend(Scenario.discover_scenarios(root))
            # Also see `smarts.env.HiwayEnv`
            self._scenarios_iterator = cycle(
                Scenario.variations_for_all_scenario_roots(
                    scenario_roots, list(agent_specs.keys())
                )
            )

    def generate_logs(self, observation, highwayenv_score):
        ego_state = observation.ego_vehicle_state
        start = observation.ego_vehicle_state.mission.start
        goal = observation.ego_vehicle_state.mission.goal
        path = get_path_to_goal(
            goal=goal, paths=observation.waypoint_paths, start=start
        )
        closest_wp, _ = get_closest_waypoint(
            num_lookahead=100,
            goal_path=path,
            ego_position=ego_state.position,
            ego_heading=ego_state.heading,
        )
        signed_dist_from_center = closest_wp.signed_lateral_error(ego_state.position)
        lane_width = closest_wp.lane_width * 0.5
        ego_dist_center = signed_dist_from_center / lane_width

        linear_jerk = np.linalg.norm(ego_state.linear_jerk)
        angular_jerk = np.linalg.norm(ego_state.angular_jerk)

        # Distance to goal
        ego_2d_position = ego_state.position[0:2]
        goal_dist = distance.euclidean(ego_2d_position, goal.position)

        angle_error = closest_wp.relative_heading(
            ego_state.heading
        )  # relative heading radians [-pi, pi]

        # number of violations
        (ego_num_violations, social_num_violations,) = ego_social_safety(
            observation,
            d_min_ego=1.0,
            t_c_ego=1.0,
            d_min_social=1.0,
            t_c_social=1.0,
            ignore_vehicle_behind=True,
        )

        info = dict(
            position=ego_state.position,
            speed=ego_state.speed,
            steering=ego_state.steering,
            heading=ego_state.heading,
            dist_center=abs(ego_dist_center),
            start=start,
            goal=goal,
            closest_wp=closest_wp,
            events=observation.events,
            ego_num_violations=ego_num_violations,
            social_num_violations=social_num_violations,
            goal_dist=goal_dist,
            linear_jerk=np.linalg.norm(ego_state.linear_jerk),
            angular_jerk=np.linalg.norm(ego_state.angular_jerk),
            env_score=self.ultra_scores(observation, highwayenv_score),
        )
        return info

    def step(self, agent_actions):
        agent_actions = {
            agent_id: self._agent_specs[agent_id].action_adapter(action)
            for agent_id, action in agent_actions.items()
        }

        observations, rewards, agent_dones, extras = self._smarts.step(agent_actions)

        infos = {
            agent_id: {"score": value, "env_obs": observations[agent_id]}
            for agent_id, value in extras["scores"].items()
        }

        for agent_id in observations:
            agent_spec = self._agent_specs[agent_id]
            observation = observations[agent_id]
            reward = rewards[agent_id]
            info = infos[agent_id]

            rewards[agent_id] = agent_spec.reward_adapter(observation, reward)
            observations[agent_id] = agent_spec.observation_adapter(observation)
            infos[agent_id] = agent_spec.info_adapter(observation, reward, info)
            infos[agent_id]["logs"] = self.generate_logs(observation, reward)

        for done in agent_dones.values():
            self._dones_registered += 1 if done else 0

        agent_dones["__all__"] = self._dones_registered == len(self._agent_specs)

        return observations, rewards, agent_dones, infos

    def get_task(self, task_id, task_level):
        base_dir = os.path.join(os.path.dirname(__file__), "../")
        config_path = os.path.join(base_dir, "config.yaml")

        with open(config_path, "r") as task_file:
            scenarios = yaml.safe_load(task_file)["tasks"]
            task = scenarios[f"task{task_id}"][task_level]

        task["train"] = os.path.join(base_dir, task["train"])
        task["test"] = os.path.join(base_dir, task["test"])
        return task

    @property
    def info(self):
        return {
            "scenario_info": self.scenario_info,
            "timestep_sec": self.timestep_sec,
            "headless": self.headless,
        }



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import os
import shutil
import unittest

import gym
import numpy as np
import ray

from smarts.zoo.registry import make
from ultra.utils.episode import episodes

AGENT_ID = "001"
timestep_sec = 0.1
seed = 2
task_id = "00"
task_level = "easy"


class EpisodeTest(unittest.TestCase):
    # Put generated files and folders in this directory.
    OUTPUT_DIRECTORY = "tests/episode_test/"

    def test_episode_record(self):
        @ray.remote(max_calls=1, num_gpus=0, num_cpus=1)
        def run_experiment():
            log_dir = os.path.join(EpisodeTest.OUTPUT_DIRECTORY, "logs/")
            agent, env = prepare_test_env_agent()
            result = {
                "episode_reward": 0,
                "dist_center": 0,
                "goal_dist": 0,
                "speed": 0,
                "ego_num_violations": 0,
                "linear_jerk": 0,
                "angular_jerk": 0,
                "collision": 0,
                "off_road": 0,
                "off_route": 0,
                "reached_goal": 0,
            }
            for episode in episodes(1, etag="Train", log_dir=log_dir):
                observations = env.reset()
                total_step = 0
                episode.reset()
                dones, infos = {"__all__": False}, None
                state = observations[AGENT_ID]

                while not dones["__all__"] and total_step < 4:
                    action = agent.act(state, explore=True)
                    observations, rewards, dones, infos = env.step({AGENT_ID: action})
                    next_state = observations[AGENT_ID]
                    # observations[AGENT_ID]["ego"].update(rewards[AGENT_ID]["log"])
                    loss_outputs = {
                        AGENT_ID: agent.step(
                            state=state,
                            action=action,
                            reward=rewards[AGENT_ID],
                            next_state=next_state,
                            done=dones[AGENT_ID],
                            info=infos[AGENT_ID],
                        )
                    }

                    for key in result.keys():
                        if key in observations[AGENT_ID]:
                            if key == "goal_dist":
                                result[key] = observations[AGENT_ID]
                            else:
                                result[key] += observations[AGENT_ID][key]
                        elif key == "episode_reward":
                            result[key] += rewards[AGENT_ID]

                    episode.record_step(
                        agent_ids_to_record=[AGENT_ID],
                        infos=infos,
                        rewards=rewards,
                        total_step=total_step,
                        loss_outputs=loss_outputs,
                    )

                    state = next_state
                    total_step += 1
            env.close()
            episode.record_episode()
            return result, episode.info

        ray.init(ignore_reinit_error=True)
        result, episode_info = ray.get(run_experiment.remote())
        for key in result.keys():
            self.assertTrue(True)
            # if key in ["episode_reward", "goal_dist"]:
            #     print(abs(result[key] - episode_info["Train"][AGENT_ID].data[key]) <= 0.001)
            #     # self.assertTrue(
            #     #     abs(result[key] - episode_info["Train"][AGENT_ID].data[key]) <= 0.001
            #     # )

    @unittest.skip
    def test_episode_counter(self):
        @ray.remote(max_calls=1, num_gpus=0, num_cpus=1)
        def run_experiment():
            agent, env = prepare_test_env_agent()
            episode_count = 0
            log_dir = os.path.join(EpisodeTest.OUTPUT_DIRECTORY, "logs/")
            for episode in episodes(2, etag="Train", log_dir=log_dir):
                observations = env.reset()
                total_step = 0
                episode.reset()
                dones, infos = {"__all__": False}, None
                state = observations[AGENT_ID]
                while not dones["__all__"]:
                    action = agent.act(state, explore=True)
                    observations, rewards, dones, infos = env.step({AGENT_ID: action})
                    next_state = observations[AGENT_ID]
                    # observations[AGENT_ID].update(rewards[AGENT_ID])
                    loss_output = agent.step(
                        state=state,
                        action=action,
                        reward=rewards[AGENT_ID],
                        next_state=next_state,
                        done=dones[AGENT_ID],
                        info=infos[AGENT_ID],
                    )
                    episode.record_step(
                        agent_id=AGENT_ID,
                        infos=infos,
                        rewards=rewards,
                        total_step=total_step,
                        loss_output=loss_output,
                    )
                    state = next_state
                    total_step += 1
                episode_count += 1
            env.close()
            return episode_count

        ray.init(ignore_reinit_error=True)
        episode_count = ray.get(run_experiment.remote())
        ray.shutdown()
        self.assertTrue(episode_count == 2)

    # def test_tensorboard_handle(self):
    #     # TODO test numbers
    #     self.assertTrue(True)

    # def test_save_model(self):
    #     self.assertTrue(True)

    # def test_save_code(self):
    #     self.assertTrue(True)

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(EpisodeTest.OUTPUT_DIRECTORY):
            shutil.rmtree(EpisodeTest.OUTPUT_DIRECTORY)


def prepare_test_env_agent(headless=True):
    timestep_sec = 0.1
    # [throttle, brake, steering]
    policy_class = "ultra.baselines.ppo:ppo-v0"
    spec = make(locator=policy_class)
    env = gym.make(
        "ultra.env:ultra-v0",
        agent_specs={AGENT_ID: spec},
        scenario_info=("00", "easy"),
        headless=headless,
        timestep_sec=timestep_sec,
        seed=seed,
    )
    agent = spec.build_agent()
    return agent, env



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import argparse
import copy
import itertools
import json
import math
import os
import random
import shutil
import time
from collections import Counter, defaultdict
from dataclasses import replace
from multiprocessing import Manager, Process
from shutil import copyfile

import numpy as np
import yaml

from smarts.sstudio import gen_missions, gen_traffic
from smarts.sstudio.types import (
    Distribution,
    Flow,
    MapZone,
    Mission,
    Route,
    Traffic,
    TrapEntryTactic,
)
from ultra.scenarios.common.begin_time_init_funcs import *
from ultra.scenarios.common.distributions import get_pattern
from ultra.scenarios.common.social_vehicle_definitions import (
    get_social_vehicle_behavior,
)

LANE_LENGTH = 137.85


def copy_map_files(scenario, map_dir, speed):
    if not os.path.exists(scenario):
        os.makedirs(scenario)
    copyfile(map_dir + f"/{speed}/map.net.xml", scenario + "/map.net.xml")
    copyfile(map_dir + f"/{speed}/map.glb", scenario + "/map.glb")


def get_direction(route_key):
    route_direction = {
        "north-south": ["north-NS", "south-NS"],
        "south-north": ["south-SN", "north-SN"],
        "south-east": ["south-SN", "east-WE"],
        "south-west": ["south-SN", "west-EW"],
        "north-east": ["north-NS", "east-WE"],
        "north-west": ["north-NS", "west-EW"],
        "east-north": ["east-EW", "north-SN"],
        "east-south": ["east-EW", "south-NS"],
        "east-west": ["east-EW", "west-EW"],
        "west-north": ["west-WE", "north-SN"],
        "west-south": ["west-WE", "south-NS"],
        "west-east": ["west-WE", "east-WE"],
    }
    return route_direction[route_key]


def generate_stopwatcher(
    stopwatcher_behavior, stopwatcher_route, start_lane_id, end_lane_id, begin_time=None
):
    # this is a vehicle with aggressive behavior that records the number of steps to finishe a left-turn
    behavior = replace(
        get_social_vehicle_behavior(stopwatcher_behavior),
        name=f"stopwatcher_{stopwatcher_behavior}",
    )
    begin_time = (
        begin_time if begin_time is not None else random.randint(10, 15)
    )  # slight delay to wait for trafficne_id
    return Flow(
        begin=begin_time,
        end=begin_time + 3600,  # 1 hour
        route=Route(
            begin=(
                f"edge-{stopwatcher_route[0]}",
                start_lane_id,
                "base",
            ),
            end=(
                f"edge-{stopwatcher_route[1]}",
                end_lane_id,
                "max",
            ),
        ),
        rate=1,
        actors={behavior: 1.0},
    )


def generate_left_turn_missions(
    missions,
    route_distributions,
    route_lanes,
    speed,
    map_dir,
    level_name,
    save_dir,
    stopwatcher_behavior,
    stopwatcher_route,
    seed,
    intersection_name,
    traffic_density,
):
    # dont worry about these seeds, theyre used by sumo
    sumo_seed = random.choice([0, 1, 2, 3, 4])
    stopwatcher_info = None
    stopwatcher_added = False
    if stopwatcher_behavior:
        stopwatcher_info = {
            "behavior": stopwatcher_behavior,
            "direction": get_direction(stopwatcher_route),
        }

    random.seed(seed)
    np.random.seed(seed)
    all_flows = []
    metadata = {"routes": {}, "total_vehicles": 0, "stopwatcher": None}

    scenario = save_dir + f"-flow-{seed}"

    # Remove old traffic route if it exists(otherwise, won't update to new flows)
    if os.path.exists(f"{scenario}/traffic/all.rou.xml"):
        shutil.rmtree(scenario)

    for route_key, route_info in route_distributions["routes"].items():
        # to skip None
        if route_info:
            if stopwatcher_behavior:  # put the ego on the side road
                ego_routes = [
                    Route(
                        begin=(
                            "edge-south-side",  # Edge
                            0,  # Lane index
                            100,  # Offset
                        ),
                        end=(
                            "edge-dead-end",
                            0,
                            5,
                        ),  # Edge  # Lane index  # Offset
                    )
                ]
            else:
                ego_routes = [
                    Route(
                        begin=(
                            "edge-{}".format(mission["start"]),  # Edge
                            route_lanes[mission["start"]] - 1,  # Lane index
                            random.randint(50, 120),  # Offset
                        ),
                        end=(
                            "edge-{}".format(mission["end"]),  # Edge
                            route_lanes[mission["end"]] - 1,  # Lane index
                            random.randint(50, 120),  # Offset
                        ),
                    )
                    for mission in missions
                ]

            flows, vehicles_log_info = generate_social_vehicles(
                route_distribution=route_info["distribution"],
                begin_time_init=route_info["begin_time_init"],
                num_vehicles=route_info["vehicles"],
                route_direction=get_direction(route_key),
                route_lanes=route_lanes,
                route_has_turn=route_info["has_turn"],
                start_end_on_different_lanes_probability=route_info[
                    "start_end_on_different_lanes_probability"
                ],
                deadlock_optimization=route_info["deadlock_optimization"],
                stopwatcher_info=stopwatcher_info,
            )
            if (
                stopwatcher_behavior
                and len(flows) > 0
                and get_direction(route_key) == stopwatcher_info["direction"]
            ):
                stopwatcher_added = True
                print(
                    f'stop watcher added to {get_direction(route_key)} flows among {route_info["vehicles"]} vehicles!'
                )
            all_flows.extend(flows)
            metadata["routes"][route_key] = vehicles_log_info
    scenario = save_dir + f"-flow-{seed}"
    if stopwatcher_behavior:
        if not stopwatcher_added and level_name != "no-traffic":
            print(
                f'There was no matching flows for stopwatcher, adding it to {stopwatcher_info["direction"]}.'
            )
            # vehicles_log_info[f'stopwatcher_{stopwatcher_info["behavior"]}']["count"] += 1
            all_flows.append(
                generate_stopwatcher(
                    stopwatcher_behavior=stopwatcher_info["behavior"],
                    stopwatcher_route=stopwatcher_info["direction"],
                    start_lane_id=0,
                    end_lane_id=0,
                )
            )
        scenario += f"-stopwatcher-{stopwatcher_info['behavior']}"

    copy_map_files(scenario, map_dir, speed)
    if stopwatcher_behavior or "ego_hijacking_params" not in route_distributions:
        mission_objects = [Mission(ego_route) for ego_route in ego_routes]
    else:
        speed_m_per_s = float("".join(filter(str.isdigit, speed))) * 5.0 / 18.0
        hijacking_params = route_distributions["ego_hijacking_params"]
        zone_range = hijacking_params["zone_range"]
        waiting_time = hijacking_params["wait_to_hijack_limit_s"]
        start_time = (
            hijacking_params["start_time"]
            if hijacking_params["start_time"] != "default"
            else random.randint((LANE_LENGTH // speed_m_per_s), 60)
        )
        mission_objects = [
            Mission(
                ego_route,
                # Optional: control hijacking time, place, and emission.
                start_time=start_time,  # When to start hijacking (might start later).
                entry_tactic=TrapEntryTactic(
                    wait_to_hijack_limit_s=waiting_time,  # When to give up on hijacking and start emitting a social vehicle instead.
                    zone=MapZone(
                        start=(
                            ego_route.begin[0],
                            0,
                            ego_route.begin[2] + zone_range[0],
                        ),
                        length=zone_range[1],
                        n_lanes=(ego_route.begin[1] + 1),
                    ),  # Area to hijack.
                    exclusion_prefixes=tuple(),  # Vehicles to be excluded (check vehicle IDs).
                ),
            )
            for ego_route in ego_routes
        ]
    random.shuffle(
        mission_objects
    )  # Shuffle the missions so agents don't do the same route all the time.
    gen_missions(scenario, mission_objects)

    traffic = Traffic(flows=all_flows)
    gen_traffic(scenario, traffic, name=f"all", seed=sumo_seed)
    # patch: remove route files from traffic folder to make intersection empty
    if traffic_density == "no-traffic":
        os.remove(f"{scenario}/traffic/all.rou.xml")
    if stopwatcher_behavior:
        metadata["stopwatcher"] = {
            "direction": stopwatcher_info["direction"],
            "behavior": stopwatcher_info["behavior"],
        }
    metadata["intersection"] = {
        "type": intersection_name[-1],
        "name": intersection_name,
    }
    metadata["total_vehicles"] = len(all_flows)
    metadata["seed"] = seed
    metadata["flow_id"] = f"flow_{seed}"
    with open(f"{scenario}/metadata.json", "w") as log:
        json.dump(metadata, log)
    # print(f"Finished:{scenario}")


def generate_social_vehicles(
    route_distribution,
    start_end_on_different_lanes_probability,
    num_vehicles,
    route_direction,
    route_lanes,
    route_has_turn,
    stopwatcher_info,
    begin_time_init=None,
    deadlock_optimization=True,
):
    flows = []
    behaviors = []
    log_info = {
        "num_vehicles": 0,
        "route_distribution": None,
        "start_end_on_different_lanes_probability": 0.0,
    }
    stopwatcher_added = False
    # populate random routes based on their probability
    start_lane, end_lane = route_direction
    if stopwatcher_info and route_direction == stopwatcher_info["direction"]:
        if stopwatcher_info["behavior"] not in behaviors:
            behaviors.append(f'stopwatcher_{stopwatcher_info["behavior"]}')
            num_vehicles = max(0, num_vehicles - 1)  # used 1 for stopwatcher
            stopwatcher_added = True

    behaviors.extend(
        random.choices(
            list(route_distribution.keys()),
            weights=list(route_distribution.values()),
            k=num_vehicles,
        )
    )

    random.shuffle(behaviors)  # because stopwatcher is always at the beginning

    if begin_time_init is None:
        begin_time_init_func = basic_begin_time_init_func
        begin_time_init_params = {"probability": 0.0}
    else:
        begin_time_init_func = begin_time_init["func"]
        begin_time_init_params = begin_time_init["params"]

    begin_time_init_params = (
        {} if begin_time_init_params is None else begin_time_init_params
    )
    begin_times = begin_time_init_func(
        route_lanes[start_lane], len(behaviors), **begin_time_init_params
    )
    begin_time_idx = [0 for _ in range(route_lanes[start_lane])]

    for behavior_idx in behaviors:
        if behavior_idx not in log_info:
            log_info[behavior_idx] = {"count": 0, "start_end_different_lanes": 0}
        if deadlock_optimization and route_has_turn:
            # if route has a turn, start on the left-most lane
            start_lane_id = route_lanes[start_lane] - 1
            lane_changing_options = [
                lane_id for lane_id in range(route_lanes[end_lane])
            ]
            if (
                len(lane_changing_options) > 0
                or random.uniform(0, 1) < start_end_on_different_lanes_probability
            ):
                end_lane_id = random.choice(lane_changing_options)
            else:
                end_lane_id = start_lane_id

        else:
            start_lane_id = random.randint(0, route_lanes[start_lane] - 1)
            end_lane_id = random.randint(0, route_lanes[end_lane] - 1)

        # set begin/end time
        begin_time = begin_times[start_lane_id][begin_time_idx[start_lane_id]]
        begin_time_idx[start_lane_id] += 1
        end_time = begin_time + 3600  # 1 hour
        if "stopwatcher" in behavior_idx:
            start_lane_id = route_lanes[stopwatcher_info["direction"][0]] - 1
            end_lane_id = route_lanes[stopwatcher_info["direction"][1]] - 1

            flows.append(
                generate_stopwatcher(
                    stopwatcher_behavior=stopwatcher_info["behavior"],
                    stopwatcher_route=stopwatcher_info["direction"],
                    begin_time=begin_time,
                    start_lane_id=start_lane_id,
                    end_lane_id=end_lane_id,
                )
            )
        else:
            behavior = get_social_vehicle_behavior(behavior_idx)
            flows.append(
                Flow(
                    begin=begin_time,
                    end=end_time,
                    route=Route(
                        begin=(f"edge-{start_lane}", start_lane_id, "base"),
                        end=(f"edge-{end_lane}", end_lane_id, "max"),
                    ),
                    rate=1,
                    actors={behavior: 1.0},
                )
            )

        log_info[behavior_idx]["count"] += 1
        log_info["route_distribution"] = route_distribution
        log_info["num_vehicles"] = (
            num_vehicles + 1 if stopwatcher_added else num_vehicles
        )
        log_info[
            "start_end_on_different_lanes_probability"
        ] = start_end_on_different_lanes_probability
        log_info[behavior_idx]["start_end_different_lanes"] += (
            1 if start_lane_id != end_lane_id else 0
        )

    return flows, log_info


def scenario_worker(
    seeds,
    ego_missions,
    route_lanes,
    route_distributions,
    map_dir,
    stopwatcher_route,
    level_name,
    save_dir,
    speed,
    stopwatcher_behavior,
    traffic_density,
    intersection_type,
    mode,
    total_seeds,
    percent,
    dynamic_pattern_func,
):
    for i, seed in enumerate(seeds):
        if not dynamic_pattern_func is None:
            route_distributions = dynamic_pattern_func(route_distributions, i)

        generate_left_turn_missions(
            missions=ego_missions,
            route_lanes=route_lanes,
            route_distributions=route_distributions,
            map_dir=map_dir,
            level_name=level_name,
            save_dir=save_dir,
            speed=speed,
            stopwatcher_behavior=stopwatcher_behavior,
            stopwatcher_route=stopwatcher_route,
            seed=seed,
            traffic_density=traffic_density,
            intersection_name=intersection_type,
        )
    # print(
    #     f"{mode} {intersection_type} {speed} {traffic_density}, counts:{len(seeds)}, generated:{len(seeds)/len(total_seeds)}, real:{percent}"
    # )


def build_scenarios(
    task,
    level_name,
    stopwatcher_behavior,
    stopwatcher_route,
    save_dir,
    root_path,
    dynamic_pattern_func=None,
):
    print("Generating Scenario ...")
    manager = Manager()

    log_dict = manager.dict()

    with open(f"{root_path}/{task}/config.yaml", "r") as task_file:
        task_config = yaml.safe_load(task_file)
        print(f"{root_path}/{task}/config.yaml")

    level_config = task_config["levels"][level_name]
    scenarios_dir = os.path.dirname(os.path.realpath(__file__))
    task_dir = f"{scenarios_dir}/{task}"
    pool_dir = f"{scenarios_dir}/pool"

    train_total, test_total = (
        int(level_config["train"]["total"]),
        int(level_config["test"]["total"]),
    )
    splitted_seeds = {
        "train": [i for i in range(train_total)],
        "test": [i for i in range(train_total, train_total + test_total)],
    }
    jobs = []
    # print(M)
    start = time.time()
    for mode, mode_seeds in splitted_seeds.items():
        combinations = []

        # Obtain the ego missions specified for this mode and ensure
        # that test scenarios only have one ego mission.
        ego_missions = level_config[mode]["ego_missions"]
        assert not (
            mode == "test" and (len(ego_missions) != 1)
        ), "Test scenarios must have one ego mission."

        prev_split = 0
        main_seed_count = 0
        # sort inverse by percents
        intersection_types = level_config[mode]["intersection_types"]
        intersections = sorted(
            [
                [_type, intersection_types[_type]["percent"]]
                for _type in intersection_types
            ],
            key=lambda x: x[1],
            reverse=True,
        )
        log_dict[mode] = {x: {"count": 0, "percent": 0} for x in intersection_types}
        for intersection_type, intersection_percent in intersections:
            part = int(float(intersection_percent) * len(mode_seeds))
            cur_split = prev_split + part
            seeds = mode_seeds[prev_split:cur_split]
            specs = sorted(
                intersection_types[intersection_type]["specs"],
                key=lambda x: float(x[2]),
                reverse=True,
            )
            seed_count = 0
            map_dir = f"{pool_dir}/{intersection_type}"
            with open(f"{map_dir}/info.json") as jsonfile:
                map_metadata = json.load(jsonfile)
                route_lanes = map_metadata["num_lanes"]
            inner_prev_split = 0
            for speed, traffic_density, percent in specs:
                inner_part = math.ceil(float(percent) * len(seeds))

                inner_cur_split = inner_prev_split + inner_part
                name_additions = [mode, level_name, intersection_type, speed]

                if level_name != "no-traffic":
                    name_additions.append(traffic_density)

                route_distributions = get_pattern(traffic_density, intersection_type)
                temp_seeds = seeds[inner_prev_split:inner_cur_split]
                seed_count += len(temp_seeds)
                if save_dir is None:
                    temp_save_dir = task_dir + "/" + "_".join(name_additions)
                else:
                    temp_save_dir = save_dir

                sub_proc = Process(
                    target=scenario_worker,
                    args=(
                        temp_seeds,
                        ego_missions,
                        route_lanes,
                        route_distributions,
                        map_dir,
                        stopwatcher_route,
                        level_name,
                        temp_save_dir,
                        speed,
                        stopwatcher_behavior,
                        traffic_density,
                        intersection_type,
                        mode,
                        seeds,
                        percent,
                        dynamic_pattern_func,
                    ),
                )
                jobs.append(sub_proc)
                sub_proc.start()
                inner_prev_split = inner_cur_split
            print(
                f">> {mode} {intersection_type} count:{seed_count} generated: {seed_count/len(mode_seeds)} real: {intersection_percent}"
            )
            # print("--")
            prev_split = cur_split
            main_seed_count += seed_count
        # print(f"Finished: {mode}  {main_seed_count/(train_total+test_total)}")
        # print("--------------------------------------------")
    for process in jobs:
        process.join()
    print("*** time took:", time.time() - start)


if __name__ == "__main__":
    parser = argparse.ArgumentParser("generate scenarios")
    parser.add_argument("--task", help="type a task id [0, 1, 2, 3]", type=str)
    parser.add_argument("--level", help="easy/medium/hard, lo-hi/hi-lo", type=str)
    parser.add_argument(
        "--stopwatcher",
        help="all/aggressive/default/slow/blocker/crusher south-west",
        nargs="+",
    )
    parser.add_argument(
        "--save-dir", help="directory for saving maps", type=str, default=None
    )
    parser.add_argument(
        "--root-dir", help="directory of maps", type=str, default="ultra/scenarios"
    )

    args = parser.parse_args()

    stopwatcher_behavior, stopwatcher_route = None, None
    if args.stopwatcher:
        stopwatcher_behavior, stopwatcher_route = args.stopwatcher

    print("starting ...")
    build_scenarios(
        task=f"task{args.task}",
        level_name=args.level,
        stopwatcher_behavior=stopwatcher_behavior,
        stopwatcher_route=stopwatcher_route,
        save_dir=args.save_dir,
        root_path=args.root_dir,
    )



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import torch
from torch import nn
import numpy as np
from ultra.baselines.bdqn.bdqn.network import *
from smarts.core.agent import Agent
from ultra.utils.common import merge_discrete_action_spaces, to_3d_action, to_2d_action
import pathlib, os, copy
from ultra.baselines.dqn.dqn.policy import DQNPolicy
from ultra.baselines.bdqn.bdqn.network import DQNWithSocialEncoder
from ultra.baselines.bdqn.bdqn.explore import EpsilonExplore
from ultra.baselines.common.replay_buffer import ReplayBuffer
from ultra.baselines.common.social_vehicle_config import get_social_vehicle_configs
from ultra.baselines.common.yaml_loader import load_yaml
from ultra.baselines.common.baseline_state_preprocessor import BaselineStatePreprocessor


class BehavioralDQNPolicy(DQNPolicy):
    def __init__(
        self,
        policy_params=None,
        checkpoint_dir=None,
    ):
        self.policy_params = policy_params
        network_class = DQNWithSocialEncoder
        self.epsilon_obj = EpsilonExplore(1.0, 0.05, 100000)

        discrete_action_spaces = [[0], [1]]
        action_size = discrete_action_spaces
        self.merge_action_spaces = -1

        self.step_count = 0
        self.update_count = 0
        self.num_updates = 0
        self.current_sticky = 0
        self.current_iteration = 0

        lr = float(policy_params["lr"])
        seed = int(policy_params["seed"])
        self.train_step = int(policy_params["train_step"])
        self.target_update = float(policy_params["target_update"])
        self.device_name = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.device = torch.device(self.device_name)
        self.warmup = int(policy_params["warmup"])
        self.gamma = float(policy_params["gamma"])
        self.batch_size = int(policy_params["batch_size"])
        self.use_ddqn = policy_params["use_ddqn"]
        self.sticky_actions = int(policy_params["sticky_actions"])
        prev_action_size = 1
        self.prev_action = np.zeros(prev_action_size)
        self.action_size = prev_action_size
        index_to_actions = [
            e.tolist() if not isinstance(e, list) else e for e in action_size
        ]

        action_to_indexs = {
            str(k): v
            for k, v in zip(
                index_to_actions, np.arange(len(index_to_actions)).astype(np.int)
            )
        }
        self.index2actions, self.action2indexs = (
            [index_to_actions],
            [action_to_indexs],
        )
        self.num_actions = [len(index_to_actions)]

        # state preprocessing
        self.social_policy_hidden_units = int(
            policy_params["social_vehicles"].get("social_policy_hidden_units", 0)
        )
        self.social_capacity = int(
            policy_params["social_vehicles"].get("social_capacity", 0)
        )
        self.observation_num_lookahead = int(
            policy_params.get("observation_num_lookahead", 0)
        )
        self.social_policy_init_std = int(
            policy_params["social_vehicles"].get("social_policy_init_std", 0)
        )
        self.num_social_features = int(
            policy_params["social_vehicles"].get("num_social_features", 0)
        )
        self.social_vehicle_config = get_social_vehicle_configs(
            **policy_params["social_vehicles"]
        )

        self.social_vehicle_encoder = self.social_vehicle_config["encoder"]
        self.state_description = BaselineStatePreprocessor.get_state_description(
            policy_params["social_vehicles"],
            policy_params["observation_num_lookahead"],
            prev_action_size,
        )
        self.social_feature_encoder_class = self.social_vehicle_encoder[
            "social_feature_encoder_class"
        ]
        self.social_feature_encoder_params = self.social_vehicle_encoder[
            "social_feature_encoder_params"
        ]

        self.checkpoint_dir = checkpoint_dir
        self.reset()

        torch.manual_seed(seed)
        network_params = {
            "state_size": self.state_size,
            "social_feature_encoder_class": self.social_feature_encoder_class,
            "social_feature_encoder_params": self.social_feature_encoder_params,
        }
        self.online_q_network = network_class(
            num_actions=self.num_actions,
            **(network_params if network_params else {}),
        ).to(self.device)
        self.target_q_network = network_class(
            num_actions=self.num_actions,
            **(network_params if network_params else {}),
        ).to(self.device)
        self.update_target_network()

        self.optimizers = torch.optim.Adam(
            params=self.online_q_network.parameters(), lr=lr
        )
        self.loss_func = nn.MSELoss(reduction="none")

        if self.checkpoint_dir:
            self.load(self.checkpoint_dir)

        self.action_space_type = "lane"
        self.to_real_action = lambda action: self.lane_actions[action[0]]
        self.replay = ReplayBuffer(
            buffer_size=int(policy_params["replay_buffer"]["buffer_size"]),
            batch_size=int(policy_params["replay_buffer"]["batch_size"]),
            device_name=self.device_name,
        )



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import torch
import torch.nn.functional as F
import numpy as np
import torch.nn as nn
import torch.nn.init as init
import torch.optim as optim


def init_fanin(tensor):
    fanin = tensor.size(1)
    v = 1.0 / np.sqrt(fanin)
    init.uniform_(tensor, -v, v)


def weights_init_(m):
    classname = m.__class__.__name__
    if classname.find("Linear") != -1:
        torch.nn.init.xavier_uniform_(m.weight, gain=1)
        torch.nn.init.constant_(m.bias, 0)


class ActorNetwork(nn.Module):
    def __init__(self, state_space, action_space, seed, social_feature_encoder=None):
        super(ActorNetwork, self).__init__()
        self.social_feature_encoder = social_feature_encoder
        self.seed = torch.manual_seed(seed)

        self.l1 = nn.Linear(state_space, 64)
        self.l2 = nn.Linear(64, 64)
        self.l3 = nn.Linear(64, action_space)
        self.reset_parameters()

    def reset_parameters(self):
        init_fanin(self.l1.weight)
        init_fanin(self.l2.weight)
        init.uniform_(self.l3.weight, -3e-3, 3e-3)

    def forward(self, state, training=False):
        low_dim_state = state["low_dim_states"]
        social_vehicles_state = state["social_vehicles"]

        aux_losses = {}
        social_feature = []
        if self.social_feature_encoder is not None:
            social_feature, social_encoder_aux_losses = self.social_feature_encoder(
                social_vehicles_state, training
            )
            aux_losses.update(social_encoder_aux_losses)
        else:
            social_feature = [e.reshape(1, -1) for e in social_vehicles_state]

        social_feature = torch.cat(social_feature, 0) if len(social_feature) > 0 else []

        state = (
            torch.cat([low_dim_state, social_feature], -1)
            if len(social_feature) > 0
            else low_dim_state
        )

        # print('*** ACTOR STATE SIZE', state.shape)
        a = F.relu(self.l1(state))
        a = F.relu(self.l2(a))
        a = torch.tanh(self.l3(a))
        if training:
            return a, aux_losses
        else:
            return a


class CriticNetwork(nn.Module):
    def __init__(self, state_space, action_space, seed, social_feature_encoder=None):
        super(CriticNetwork, self).__init__()
        self.social_feature_encoder = social_feature_encoder
        self.seed = torch.manual_seed(seed)

        self.l1 = nn.Linear(state_space, 64)
        self.l2 = nn.Linear(64 + action_space, 64)
        self.l3 = nn.Linear(64, 1)
        self.reset_parameters()

    def reset_parameters(self):
        init_fanin(self.l1.weight)
        init_fanin(self.l2.weight)
        init_fanin(self.l3.weight)

    def forward(self, state, action, training=False):
        low_dim_state = state["low_dim_states"]
        social_vehicles_state = state["social_vehicles"]

        aux_losses = {}
        social_feature = []
        if self.social_feature_encoder is not None:
            social_feature, social_encoder_aux_losses = self.social_feature_encoder(
                social_vehicles_state, training
            )
            aux_losses.update(social_encoder_aux_losses)
        else:
            social_feature = [e.reshape(1, -1) for e in social_vehicles_state]

        social_feature = torch.cat(social_feature, 0) if len(social_feature) > 0 else []

        state = (
            torch.cat([low_dim_state, social_feature], -1)
            if len(social_feature) > 0
            else low_dim_state
        )

        # print('*** CRITIC STATE SIZE', state.shape)
        q = F.relu(self.l1(state))
        q = F.relu(self.l2(torch.cat([q, action], 1)))
        q = self.l3(q)
        if training:
            return q, aux_losses
        else:
            return q



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import os
import pickle

# Set environment to better support Ray
os.environ["MKL_NUM_THREADS"] = "1"
import argparse
import glob
import time
from pydoc import locate

import gym
import numpy as np
import ray
import torch

from smarts.zoo.registry import make
from ultra.utils.episode import LogInfo, episodes
from ultra.utils.ray import default_ray_kwargs

num_gpus = 1 if torch.cuda.is_available() else 0


def evaluation_check(
    agents,
    episode,
    agent_ids,
    policy_classes,
    eval_rate,
    eval_episodes,
    max_episode_steps,
    scenario_info,
    timestep_sec,
    headless,
    log_dir,
):
    # Evaluate agents that have reached the eval_rate.
    agent_ids_to_evaluate = [
        agent_id
        for agent_id in agent_ids
        if (episode.get_itr(agent_id) + 1) % eval_rate == 0
        and episode.last_eval_iterations[agent_id] != episode.get_itr(agent_id)
    ]

    # Skip evaluation if there are no agents needing an evaluation.
    if len(agent_ids_to_evaluate) < 1:
        return

    episode.eval_mode()
    evaluation_data = {}

    for agent_id in agent_ids_to_evaluate:
        # Get the checkpoint directory for the current agent and save its model.
        checkpoint_directory = episode.checkpoint_dir(
            agent_id, episode.get_itr(agent_id)
        )
        agents[agent_id].save(checkpoint_directory)

        # Perform the evaluation on this agent and save the data.
        evaluation_data.update(
            ray.get(
                [
                    evaluate.remote(
                        seed=episode.eval_count,
                        experiment_dir=episode.experiment_dir,
                        agent_ids=[agent_id],
                        policy_classes={agent_id: policy_classes[agent_id]},
                        checkpoint_dirs={agent_id: checkpoint_directory},
                        scenario_info=scenario_info,
                        num_episodes=eval_episodes,
                        max_episode_steps=max_episode_steps,
                        headless=headless,
                        timestep_sec=timestep_sec,
                        log_dir=log_dir,
                    )
                ]
            )[0]
        )
        episode.eval_count += 1
        episode.last_eval_iterations[agent_id] = episode.get_itr(agent_id)

    # Put the evaluation data for all agents into the episode and record the TensorBoard.
    episode.info[episode.active_tag] = evaluation_data
    episode.record_tensorboard()
    episode.train_mode()


# Number of GPUs should be split between remote functions.
@ray.remote(num_gpus=num_gpus / 2)
def evaluate(
    experiment_dir,
    seed,
    agent_ids,
    policy_classes,
    checkpoint_dirs,
    scenario_info,
    num_episodes,
    max_episode_steps,
    headless,
    timestep_sec,
    log_dir,
):
    torch.set_num_threads(1)

    # Create the agent specifications matched with their associated ID.
    agent_specs = {
        agent_id: make(
            locator=policy_classes[agent_id],
            checkpoint_dir=checkpoint_dirs[agent_id],
            experiment_dir=experiment_dir,
            max_episode_steps=max_episode_steps,
            agent_id=agent_id,
        )
        for agent_id in agent_ids
    }

    # Create the environment with the specified agents.
    env = gym.make(
        "ultra.env:ultra-v0",
        agent_specs=agent_specs,
        scenario_info=scenario_info,
        headless=headless,
        timestep_sec=timestep_sec,
        seed=seed,
        eval_mode=True,
    )

    # Build each agent from its specification.
    agents = {
        agent_id: agent_spec.build_agent()
        for agent_id, agent_spec in agent_specs.items()
    }

    # A dictionary to hold the evaluation data for each agent.
    summary_log = {agent_id: LogInfo() for agent_id in agent_ids}

    # Define an 'etag' for this experiment's data directory based off policy_classes.
    # E.g. From a ["ultra.baselines.dqn:dqn-v0", "ultra.baselines.ppo:ppo-v0"]
    # policy_classes list, transform it to an etag of "dqn-v0:ppo-v0".
    etag = ":".join([policy_class.split(":")[-1] for policy_class in policy_classes])

    for episode in episodes(num_episodes, etag=etag, log_dir=log_dir):
        # Reset the environment and retrieve the initial observations.
        observations = env.reset()
        dones = {"__all__": False}
        infos = None
        episode.reset(mode="Evaluation")

        while not dones["__all__"]:
            # Get and perform the available agents' actions.
            actions = {
                agent_id: agents[agent_id].act(observation, explore=False)
                for agent_id, observation in observations.items()
            }
            observations, rewards, dones, infos = env.step(actions)

            # Record the data from this episode.
            episode.record_step(
                agent_ids_to_record=infos.keys(), infos=infos, rewards=rewards
            )

        episode.record_episode()

        for agent_id, agent_data in episode.info[episode.active_tag].items():
            for key, value in agent_data.data.items():
                if not isinstance(value, (list, tuple, np.ndarray)):
                    summary_log[agent_id].data[key] += value

    # Normalize by the number of evaluation episodes.
    for agent_id, agent_data in summary_log.items():
        for key, value in agent_data.data.items():
            if not isinstance(value, (list, tuple, np.ndarray)):
                summary_log[agent_id].data[key] /= num_episodes

    env.close()

    return summary_log


if __name__ == "__main__":
    parser = argparse.ArgumentParser("intersection-evaluation")
    parser.add_argument(
        "--task", help="Tasks available : [0, 1, 2]", type=str, default="1"
    )
    parser.add_argument(
        "--level",
        help="Levels available : [easy, medium, hard, no-traffic]",
        type=str,
        default="easy",
    )
    parser.add_argument(
        "--models",
        nargs="+",
        help="The models in the models/ directory of the experiment to evaluate",
    )
    parser.add_argument(
        "--episodes", help="Number of training episodes", type=int, default=200
    )
    parser.add_argument(
        "--max-episode-steps",
        help="Maximum number of steps per episode",
        type=int,
        default=10000,
    )
    parser.add_argument(
        "--timestep", help="Environment timestep (sec)", type=float, default=0.1
    )
    parser.add_argument(
        "--headless", help="Run without envision", type=bool, default=False
    )
    parser.add_argument(
        "--experiment-dir",
        help="Path to the experiment directory",
        type=str,
    )
    parser.add_argument(
        "--log-dir",
        help="Log directory location",
        default="logs",
        type=str,
    )
    args = parser.parse_args()

    if not os.path.exists(args.log_dir):
        os.makedirs(args.log_dir)

    if not all([os.path.exists(model_path) for model_path in args.models]):
        raise "At least one path to a model is invalid"

    if not all([os.listdir(model_path) for model_path in args.models]):
        raise "There are no models to evaluate in at least one model path"

    agent_ids_from_models = [
        os.path.basename(os.path.normpath(model_path)) for model_path in args.models
    ]

    # Load relevant agent metadata.
    with open(
        os.path.join(args.experiment_dir, "agent_metadata.pkl"), "rb"
    ) as metadata_file:
        agent_metadata = pickle.load(metadata_file)

    # Extract the agent IDs and policy classes from the metadata and given models.
    agent_ids = [
        agent_id
        for agent_id in agent_metadata["agent_ids"]
        if agent_id in agent_ids_from_models
    ]
    policy_classes = {
        agent_id: agent_metadata["agent_classes"][agent_id] for agent_id in agent_ids
    }

    # From a base model directory such as logs/<experiment_name>/models/*, assign each agent ID with its
    # checkpoint directories in sorted order based on the checkpoint iteration. The agent IDs are
    # obtained from the direct child folders of the model directory given. As an example result:
    # {
    #     '000': ['logs/<experiment_name>/models/000/1042', 'logs/<experiment_name>/models/000/2062'],
    #     '001': ['logs/<experiment_name>/models/001/999', 'logs/<experiment_name>/models/001/1999'],
    #     '003': ['logs/<experiment_name>/models/003/1009', 'logs/<experiment_name>/models/003/2120'],
    #     '002': ['logs/<experiment_name>/models/002/1053', 'logs/<experiment_name>/models/002/2041'],
    # }
    agent_checkpoint_directories = {
        agent_id: sorted(
            glob.glob(os.path.join(args.experiment_dir, "models", agent_id, "*")),
            key=lambda x: int(x.split("/")[-1]),
        )
        for agent_id in agent_ids
    }

    # Assert each agent ID has the same number of checkpoints saved.
    directories_iterator = iter(agent_checkpoint_directories.values())
    number_of_checkpoints = len(next(directories_iterator))
    assert all(
        len(checkpoint_directory) == number_of_checkpoints
        for checkpoint_directory in directories_iterator
    ), "Not all agents have the same number of checkpoints saved"

    # Define an 'etag' for this experiment's data directory based off policy_classes.
    # E.g. From a {"000": "ultra.baselines.dqn:dqn-v0", "001": "ultra.baselines.ppo:ppo-v0"]
    # policy_classes dict, transform it to an etag of "dqn-v0:ppo-v0-evaluation".
    etag = (
        ":".join([policy_classes[agent_id].split(":")[-1] for agent_id in agent_ids])
        + "-evaluation"
    )

    ray.init()
    try:
        for episode in episodes(
            number_of_checkpoints,
            etag=etag,
            log_dir=args.log_dir,
        ):
            # Obtain a checkpoint directory for each agent.
            current_checkpoint_directories = {
                agent_id: agent_directories[episode.index]
                for agent_id, agent_directories in agent_checkpoint_directories.items()
            }
            episode.eval_mode()
            episode.info[episode.active_tag] = ray.get(
                [
                    evaluate.remote(
                        experiment_dir=args.experiment_dir,
                        agent_ids=agent_ids,
                        policy_classes=policy_classes,
                        seed=episode.eval_count,
                        checkpoint_dirs=current_checkpoint_directories,
                        scenario_info=(args.task, args.level),
                        num_episodes=int(args.episodes),
                        max_episode_steps=int(args.max_episode_steps),
                        timestep_sec=float(args.timestep),
                        headless=args.headless,
                        log_dir=args.log_dir,
                    )
                ]
            )[0]
            episode.record_tensorboard()
            episode.eval_count += 1
    finally:
        time.sleep(1)
        ray.shutdown()



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import json
import numpy as np
import torch, yaml, os, inspect, dill
from smarts.core.controllers import ActionSpaceType
from smarts.core.agent_interface import (
    AgentInterface,
    AgentType,
    OGM,
    Waypoints,
    NeighborhoodVehicles,
)

from ultra.baselines.common.yaml_loader import load_yaml
from smarts.core.agent import AgentSpec
from ultra.baselines.adapter import BaselineAdapter


class BaselineAgentSpec(AgentSpec):
    def __init__(
        self,
        policy_class,
        action_type,
        checkpoint_dir=None,
        task=None,
        max_episode_steps=1200,
        experiment_dir=None,
    ):
        pass

    def __new__(
        self,
        policy_class,
        action_type,
        checkpoint_dir=None,
        task=None,
        max_episode_steps=1200,
        experiment_dir=None,
        agent_id="",
    ):
        if experiment_dir:
            print(
                f"Loading spec for {agent_id} from {experiment_dir}/agent_metadata.pkl"
            )
            with open(f"{experiment_dir}/agent_metadata.pkl", "rb") as metadata_file:
                agent_metadata = dill.load(metadata_file)
                spec = agent_metadata["agent_specs"][agent_id]

                new_spec = AgentSpec(
                    interface=spec.interface,
                    agent_params=dict(
                        policy_params=spec.agent_params["policy_params"],
                        checkpoint_dir=checkpoint_dir,
                    ),
                    agent_builder=spec.policy_builder,
                    observation_adapter=spec.observation_adapter,
                    reward_adapter=spec.reward_adapter,
                )

                spec = new_spec
        else:
            base_dir = os.path.join(os.path.dirname(__file__), "../")
            pool_path = os.path.join(base_dir, "agent_pool.json")

            policy_class_name = policy_class.__name__
            agent_name = None

            with open(pool_path, "r") as f:
                data = json.load(f)
                agents = data["agents"].keys()
                for agent in agents:
                    if data["agents"][agent]["class"] == policy_class_name:
                        agent_name = data["agents"][agent]["name"]
                        break

            assert agent_name != None

            adapter = BaselineAdapter(agent_name)
            spec = AgentSpec(
                interface=AgentInterface(
                    waypoints=Waypoints(lookahead=20),
                    neighborhood_vehicles=NeighborhoodVehicles(200),
                    action=action_type,
                    rgb=False,
                    max_episode_steps=max_episode_steps,
                    debug=True,
                ),
                agent_params=dict(
                    policy_params=adapter.policy_params, checkpoint_dir=checkpoint_dir
                ),
                agent_builder=policy_class,
                observation_adapter=adapter.observation_adapter,
                reward_adapter=adapter.reward_adapter,
            )
        return spec



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import numpy as np
import random
from scipy.spatial.distance import euclidean
from collections import defaultdict
from ultra.utils.common import rotate2d_vector

HORIZONTAL = 0
VERTICAL = 1
REAR = 2
FRONT = 3


def extract_social_vehicle_state_default(
    social_vehicle, ego_vehicle_pos, ego_vehicle_heading, social_vehicle_config=None
):
    social_vehicle_position = social_vehicle.position[0:2] - ego_vehicle_pos[0:2]
    social_vehicle_position_rotated = rotate2d_vector(
        social_vehicle_position, -ego_vehicle_heading
    )
    return [
        (social_vehicle_position_rotated[0]) / 100.0,
        (social_vehicle_position_rotated[1]) / 100.0,
        (social_vehicle.heading - ego_vehicle_heading) / 3.14,
        social_vehicle.speed / 30.0,
    ]


def extract_social_vehicle_state_pointnet(
    social_vehicle, ego_vehicle_pos, ego_vehicle_heading, social_vehicle_config
):
    speed_norm = 30
    heading_diff = social_vehicle.heading - ego_vehicle_heading
    # vector length is prop to the speed
    heading_vector = np.asarray([np.cos(heading_diff), np.sin(heading_diff)]) * (
        social_vehicle.speed / speed_norm
    )
    social_vehicle_position = social_vehicle.position[0:2] - ego_vehicle_pos[0:2]
    social_vehicle_position_rotated = rotate2d_vector(
        social_vehicle_position, -ego_vehicle_heading
    )
    return [
        (social_vehicle_position_rotated[0]) / 100.0,
        (social_vehicle_position_rotated[1]) / 100.0,
        heading_vector[0],
        heading_vector[1],
    ]


def get_social_vehicles_states_sorted_by_distance(
    ego_vehicle_pos,
    ego_vehicle_heading,
    neighborhood_vehicles,
    social_vehicle_config,
    extractor_func,
):
    social_vehicles = neighborhood_vehicles
    social_vehicles.sort(
        key=lambda vehicle: get_distance(vehicle, ego_vehicle_pos), reverse=False
    )
    social_vehicles = [
        extractor_func(
            social_vehicle=vehicle,
            ego_vehicle_pos=ego_vehicle_pos,
            ego_vehicle_heading=ego_vehicle_heading,
            social_vehicle_config=social_vehicle_config,
        )
        for vehicle in social_vehicles
    ]
    return social_vehicles


def get_social_vehicles_leading(
    ego_vehicle_pos,
    ego_vehicle_heading,
    neighborhood_vehicles,
    extractor_func,
    max_dist_social_vehicle,
    num_social_vehicle_per_lane,
    social_capacity,
    waypoint_paths,
):
    social_vehicles, vehicles_on_ego_lane = [], []
    vehicles_on_other_lanes = defaultdict(lambda: [])

    # find ego lane information
    ego_wps = [path[0] for path in waypoint_paths]

    ego_closest_waypoint, prev_ego_wp = None, None
    ego_closest_waypoint_min_dist = float("inf")
    for i, wp in enumerate(ego_wps):
        temp_dist = wp.dist_to(ego_vehicle_pos)
        if temp_dist < ego_closest_waypoint_min_dist:
            prev_ego_wp = ego_wps[max(i - 1, 0)]
            ego_closest_waypoint = wp
            ego_closest_waypoint_min_dist = temp_dist

    ego_lane_id = ego_closest_waypoint.lane_id
    ego_lane_index = ego_closest_waypoint.lane_index

    # if vehicle is within the max_dist consider:
    #   1- find vehicles on ego lane , and only keep the closest front and rear vehicles
    #   2- find vehicles on other lanes
    for vehicle in neighborhood_vehicles:
        dist = get_distance(vehicle, ego_vehicle_pos)
        if dist < max_dist_social_vehicle:
            if vehicle.lane_id == ego_lane_id and vehicle.lane_index == ego_lane_index:
                vehicles_on_ego_lane.append(vehicle)
            else:
                vehicles_on_other_lanes[(vehicle.lane_id, vehicle.lane_index)].append(
                    vehicle
                )

    # process vehicles on ego lane and only keep the nearest front/rear
    if len(vehicles_on_ego_lane) > 0:
        front_dist, rear_dist = float("inf"), float("inf")
        front_vehicle, rear_vehicle = None, None
        for vehicle in vehicles_on_ego_lane:
            location, _ = find_relative_location(
                vehicle.position, ego_vehicle_pos, ego_vehicle_heading
            )
            dist = get_distance(vehicle, ego_vehicle_pos)
            if location == FRONT and dist < front_dist:
                front_vehicle = vehicle
                front_dist = dist
            elif location == REAR and dist < rear_dist:
                rear_vehicle = vehicle
                rear_dist = dist
        if front_vehicle:
            social_vehicles.append(
                extractor_func(
                    social_vehicle=front_vehicle,
                    ego_vehicle_pos=ego_vehicle_pos,
                    ego_vehicle_heading=ego_vehicle_heading,
                )
            )
        if rear_vehicle:
            social_vehicles.append(
                extractor_func(
                    social_vehicle=rear_vehicle,
                    ego_vehicle_pos=ego_vehicle_pos,
                    ego_vehicle_heading=ego_vehicle_heading,
                )
            )

    # process vehicles on the other lanes:
    # sort vehicles based on their distance per lane and keep num_social_vehicle_per_lane
    for lane, vehicles in vehicles_on_other_lanes.items():
        vehicles_on_other_lanes[lane].sort(
            key=lambda vehicle: get_distance(vehicle, ego_vehicle_pos)
        )
        vehicles_on_other_lanes[lane] = vehicles_on_other_lanes[lane][
            : min(len(vehicles_on_other_lanes), num_social_vehicle_per_lane)
        ]
        for v in vehicles_on_other_lanes[lane]:
            social_vehicles.append(
                extractor_func(v, ego_vehicle_pos, ego_vehicle_heading)
            )
    # add fake social vehicles if necessary
    if len(social_vehicles) < social_capacity:
        # find a waypoint on ego path equal to max_dist
        fake_social_vehicle = make_fake_social_vehicle(
            ego_vehicle_pos, ego_vehicle_heading, prev_ego_wp, max_dist_social_vehicle
        )
        fake_social_vehicles = [
            fake_social_vehicle for i in range(social_capacity - len(social_vehicles))
        ]
        social_vehicles.extend(fake_social_vehicles)

    # Permute the social vehicles randomly
    social_vehicles_permuted = np.random.permutation(social_vehicles)
    # If we have more vehicles than required, then delete the tail
    if len(social_vehicles_permuted) > social_capacity:
        social_vehicles_permuted = social_vehicles_permuted[:social_capacity]

    return social_vehicles_permuted


def make_fake_social_vehicle(
    ego_vehicle_pos, ego_vehicle_heading, prev_ego_wp, max_dist_social_vehicle
):
    # DG: I changed to zeros because I think it'll be easier for learning
    return [0.0, 0.0, 0.0, 0.0]


"""
def make_fake_social_vehicle(ego_vehicle_pos, ego_vehicle_heading, prev_ego_wp, max_dist_social_vehicle):
    fake_social_vehicle = [
        ego_vehicle_pos[0],
        ego_vehicle_pos[1],
        0.0,
        0.0,
    ]
    location, orientation = find_relative_location(prev_ego_wp.pos, ego_vehicle_pos, ego_vehicle_heading)
    if orientation == VERTICAL:
        if location == REAR:
            fake_social_vehicle[1] -= max_dist_social_vehicle
        else:
            fake_social_vehicle[1] += max_dist_social_vehicle
    else:
        if location == REAR:
            fake_social_vehicle[0] += max_dist_social_vehicle
        else:
            fake_social_vehicle[0] -= max_dist_social_vehicle
    fake_social_vehicle[0] /= 100.0
    fake_social_vehicle[1] /= 100.0
    return fake_social_vehicle
"""


def find_relative_location(position, target_position, heading):
    location = None
    orientation = None
    x_diff = position[0] - target_position[0]
    y_diff = position[1] - target_position[1]

    # DG: Rotate to get relative location - This won't work due to assumptions in this code about orientation
    # rotated_diff = rotate2d_vector(np.array([x_diff, y_diff]), -heading)
    # x_diff, y_diff = rotated_diff[0], rotated_diff[1]

    # TODO: DG - I think there is a bug here since it assumes a road orientation.
    #   We probably need to
    #    1) compute the direction vector of the closest waypoint
    #    2) find the normal of the waypoint direction
    #    3) determine the side that the waypoint is on via a projection onto the normal
    #   We don't need orientation so we can just return FRONT versus REAR
    #   We won't know which is which (FRONT versus REAR) with this method but we can distinguish by the sign.
    #   One will be positive and one will be negative and we want to return the social vehicles in each lane
    #   with the smallest projection for each sign (positive and negative signs).
    #   Details: https://math.stackexchange.com/questions/274712/calculate-on-which-side-of-a-straight-line-is-a-given-point-located
    if abs(x_diff) > abs(y_diff):
        orientation = HORIZONTAL
        location = REAR if x_diff < 0 else FRONT
    else:
        orientation = VERTICAL
        location = FRONT if y_diff > 0 else REAR
    return location, orientation


def get_distance(social_vehicle, ego_vehicle_pos):
    return euclidean(social_vehicle.position[:2], ego_vehicle_pos[:2])



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import re

from smarts.sstudio.types import (
    Distribution,
    JunctionModel,
    LaneChangingModel,
    TrafficActor,
)

social_vehicle_colors = {
    "default": (255, 255, 255),
    "aggressive": (255, 128, 0),
    "cautious": (100, 178, 255),
    "blocker": (102, 0, 255),
    # "bus": (255, 255, 0),
    # "crusher": (255, 0, 0),
    "stopwatcher": (255, 158, 251),
}

depart_speed = "max"  # sets to max
social_vehicles = {
    "default": TrafficActor(
        name="default",
        accel=3.0,
        decel=7.5,
        tau=1.0,
        depart_speed=depart_speed,
        min_gap=Distribution(mean=2.5, sigma=0.05),
        speed=Distribution(mean=1.0, sigma=0.05),
        imperfection=Distribution(mean=0.5, sigma=0.05),
        vehicle_type="passenger",
        lane_changing_model=LaneChangingModel(),
        junction_model=JunctionModel(),
    ),
    "aggressive": TrafficActor(
        name="aggressive",
        accel=10.0,
        decel=10.0,
        tau=0.5,
        depart_speed=depart_speed,
        min_gap=Distribution(mean=1.0, sigma=0.1),
        speed=Distribution(mean=1.5, sigma=0.1),
        imperfection=Distribution(mean=0.1, sigma=0.1),
        vehicle_type="passenger",
        lane_changing_model=LaneChangingModel(
            strategic=10,
            pushy=1.0,
            overtake_right=1,
            speed_gain=200,
            time_to_impatience=0.01,
            impatience=1.0,
            lookahead_left=0.0,
        ),
        junction_model=JunctionModel(
            impatience=0.98, timegap_minor=0.1, ignore_foe_prob=0.98
        ),
    ),
    "cautious": TrafficActor(
        name="cautious",
        accel=3.0,
        decel=10.0,
        tau=1.0,
        depart_speed=depart_speed,
        min_gap=Distribution(mean=5.0, sigma=0.05),
        speed=Distribution(mean=0.4, sigma=0.1),
        imperfection=Distribution(mean=0.5, sigma=0.05),
        vehicle_type="passenger",
        lane_changing_model=LaneChangingModel(
            strategic=1,
            pushy=0.2,
            overtake_right=0,
            speed_gain=1,
            keep_right=1,
            impatience=0,
            lookahead_left=2.9,
        ),
        junction_model=JunctionModel(
            impatience=0.0, timegap_minor=1.0, ignore_foe_prob=0.0
        ),
    ),
    "blocker": TrafficActor(
        name="blocker",
        accel=3.0,
        decel=7.5,
        tau=1.0,
        depart_speed=depart_speed,
        min_gap=Distribution(mean=1.0, sigma=0.1),
        speed=Distribution(mean=0.2, sigma=0.1),
        imperfection=Distribution(mean=0.0, sigma=0.1),
        vehicle_type="passenger",
        lane_changing_model=LaneChangingModel(
            strategic=0,
            pushy=0,
            overtake_right=0,
            speed_gain=1,
            time_to_impatience=1000,
            cooperative=0.0,
            opposite=0,  # less lane changing behavior
            impatience=0.0,
            keep_right=0.0,
        ),
        junction_model=JunctionModel(
            impatience=0.0, timegap_minor=1.0, ignore_foe_prob=0.0
        ),
    ),
    "bus": TrafficActor(
        name="bus",
        accel=1.2,
        decel=4.0,
        tau=1.0,
        depart_speed=depart_speed,
        min_gap=Distribution(mean=2.5, sigma=0.05),
        speed=Distribution(mean=1.0, sigma=0.05),
        imperfection=Distribution(mean=0.5, sigma=0.05),
        vehicle_type="bus",
        lane_changing_model=LaneChangingModel(),
        junction_model=JunctionModel(),
    ),
    "truck": TrafficActor(
        name="truck",
        accel=1.3,
        decel=4.0,
        tau=1.0,
        depart_speed=depart_speed,
        min_gap=Distribution(mean=2.5, sigma=0.05),
        speed=Distribution(mean=1.0, sigma=0.05),
        imperfection=Distribution(mean=0.5, sigma=0.05),
        vehicle_type="truck",
        lane_changing_model=LaneChangingModel(),
        junction_model=JunctionModel(),
    ),
    "trailer": TrafficActor(
        name="trailer",
        accel=1.0,
        decel=4.0,
        tau=1.0,
        depart_speed=depart_speed,
        min_gap=Distribution(mean=2.5, sigma=0.05),
        speed=Distribution(mean=1.0, sigma=0.05),
        imperfection=Distribution(mean=0.5, sigma=0.05),
        vehicle_type="trailer",
        lane_changing_model=LaneChangingModel(),
        junction_model=JunctionModel(),
    ),
    "coach": TrafficActor(
        name="coach",
        accel=2.0,
        decel=4.0,
        tau=1.0,
        depart_speed=depart_speed,
        min_gap=Distribution(mean=2.5, sigma=0.05),
        speed=Distribution(mean=1.0, sigma=0.05),
        imperfection=Distribution(mean=0.5, sigma=0.05),
        vehicle_type="coach",
        lane_changing_model=LaneChangingModel(),
        junction_model=JunctionModel(),
    ),
    # "crusher": TrafficActor(
    #     name="crusher",
    #     accel=10.0,
    #     decel=1.0,
    #     tau=0.0,
    #     depart_speed=depart_speed,
    #     emergency_decel=0.0,
    #     sigma=1.0,
    #     min_gap=Distribution(mean=0.0, sigma=0),
    #     speed=Distribution(mean=2.0, sigma=0.1),
    #     imperfection=Distribution(mean=1.0, sigma=0),
    #     vehicle_type="passenger",
    #     lane_changing_model=LaneChangingModel(
    #         strategic=100,
    #         pushy=1.0,
    #         overtake_right=1,
    #         speed_gain=200,
    #         time_to_impatience=0.1,
    #         cooperative=0.0,
    #         opposite=0, # less lane changing behavior
    #         impatience=1.0
    #     ),
    #     junction_model=JunctionModel(
    #         impatience=1, timegap_minor=0.0, ignore_foe_prob=1
    #     ),
    # ),
}


def get_social_vehicle_behavior(behavior_id):
    if behavior_id in social_vehicles:
        return social_vehicles[behavior_id]
    return None


def get_all_behaviors():
    all_behaviors = sorted(list(social_vehicle_colors.keys()))
    all_behaviors.remove("stopwatcher")
    return all_behaviors


def get_social_vehicle_color(vehicle_id):
    behavior_id = vehicle_id.split("-")[0]
    if behavior_id in social_vehicle_colors.keys():
        return (behavior_id, social_vehicle_colors[behavior_id])
    else:
        # some processing:  actor-default-2244061760535854552, actor-stopwatcher_aggressive--2078631594661947501
        behavior_match = behavior_id
        stopwatcher_behavior = None

        behavior_key = behavior_match
        if "_" in behavior_match:
            behavior_key, stopwatcher_behavior = behavior_match.split("_")

        if stopwatcher_behavior:
            return (behavior_match, social_vehicle_colors[stopwatcher_behavior])
        if behavior_key in social_vehicle_colors.keys():
            return (behavior_match, social_vehicle_colors[behavior_key])

    return (f" Unknown ***** {behavior_id} *******", (255, 50, 255))



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import unittest

import gym
import ray

from smarts.core.controllers import ActionSpaceType
from smarts.zoo.registry import make
from ultra.baselines.agent_spec import BaselineAgentSpec
from ultra.baselines.ppo.ppo.policy import PPOPolicy

AGENT_ID = "001"
timestep_sec = 0.1
seed = 2
task_id = "00"
task_level = "easy"


class EnvTest(unittest.TestCase):
    def test_scenario_task(self):
        @ray.remote(max_calls=1, num_gpus=0, num_cpus=1)
        def run_experiment():
            agent, env = prepare_test_env_agent()
            task_id1, task_level1 = env.scenario_info
            env.close()
            return task_id1, task_level1

        ray.init(ignore_reinit_error=True)
        task_id1, task_level1 = ray.get(run_experiment.remote())
        ray.shutdown()
        self.assertTrue(task_id1 == task_id)
        self.assertTrue(task_level1 == task_level)

    def test_headless(self):
        @ray.remote(max_calls=1, num_gpus=0, num_cpus=1)
        def run_experiment():
            agent, env = prepare_test_env_agent(headless=False)
            headless1 = env.headless
            env.close()

            agent, env = prepare_test_env_agent(headless=True)
            headless2 = env.headless
            env.close()
            return headless1, headless2

        ray.init(ignore_reinit_error=True)
        headless1, headless2 = ray.get(run_experiment.remote())
        ray.shutdown()
        self.assertFalse(headless1)
        self.assertTrue(headless2)

    def test_timestep_sec(self):
        @ray.remote(max_calls=1, num_gpus=0)
        def run_experiment():
            agent, env = prepare_test_env_agent(headless=False)
            timestep_sec1 = env.timestep_sec
            env.close()
            return timestep_sec1

        ray.init(ignore_reinit_error=True)
        timestep_sec1 = ray.get(run_experiment.remote())
        ray.shutdown()
        self.assertTrue(timestep_sec1 == timestep_sec)


# other attributes are not set
#'action_space', 'close', 'get_task', 'headless', 'info', 'metadata', 'observation_space', 'render', 'reset', 'reward_range', 'scenario_info', 'scenario_log', 'scenarios', 'seed', 'spec', 'step', 'timestep_sec', 'unwrapped'


def prepare_test_env_agent(headless=True):
    timestep_sec = 0.1
    # [throttle, brake, steering]
    policy_class = "ultra.baselines.ppo:ppo-v0"
    spec = make(locator=policy_class)
    env = gym.make(
        "ultra.env:ultra-v0",
        agent_specs={AGENT_ID: spec},
        scenario_info=("00", "easy"),
        headless=headless,
        timestep_sec=timestep_sec,
        seed=seed,
    )
    agent = spec.build_agent()
    return agent, env



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import math


class RLlibLogInfo:
    def __init__(self):
        self.data = {
            "env_score": 0,
            "episode_reward": 0,
            "dist_center": 0,
            "goal_dist": 0,
            "speed": 0,
            "max_speed_violation": 0,
            "ego_num_violations": 0,
            "social_num_violations": 0,
            "ego_linear_jerk": 0.0,
            "ego_angular_jerk": 0.0,
            "final_pos": [0, 0],
            "start_pos": [0, 0],
            "dist_travelled": 0.0,
            "collision": 0,
            "off_road": 0,
            "off_route": 0,
            "reached_goal": 0,
            "timed_out": 0,
            "episode_length": 0,
        }

    def add(self, infos, rewards):

        self.data["env_score"] += int(infos["logs"]["env_score"])
        self.data["speed"] += infos["logs"]["speed"]
        self.data["max_speed_violation"] += (
            1 if infos["logs"]["speed"] > infos["logs"]["closest_wp"].speed_limit else 0
        )
        self.data["dist_center"] += infos["logs"]["dist_center"]
        self.data["ego_num_violations"] += int(infos["logs"]["ego_num_violations"] > 0)
        self.data["social_num_violations"] += int(
            infos["logs"]["social_num_violations"] > 0
        )
        self.data["goal_dist"] = infos["logs"]["goal_dist"]
        self.data["ego_linear_jerk"] += infos["logs"]["linear_jerk"]
        self.data["ego_angular_jerk"] += infos["logs"]["angular_jerk"]
        self.data["episode_reward"] += rewards
        self.data["final_pos"] = infos["logs"]["position"]
        self.data["start_pos"] = infos["logs"]["start"].position
        self.data["dist_travelled"] = math.sqrt(
            (self.data["final_pos"][1] - self.data["start_pos"][1]) ** 2
            + (self.data["final_pos"][0] - self.data["start_pos"][0]) ** 2
        )
        # recording termination cases
        events = infos["logs"]["events"]
        self.data["collision"] = (
            False
            if len(events.collisions) == 0 or events.collisions[0].collidee_id == 0
            else True
        )
        self.data["off_road"] = int(events.off_road)
        self.data["off_route"] = int(events.off_route)
        self.data["reached_goal"] = int(events.reached_goal)
        self.data["timed_out"] = int(events.reached_max_episode_steps)
        #

    def normalize(self, steps):
        self.data["env_score"] /= steps
        self.data["dist_center"] /= steps
        self.data["episode_length"] = steps
        self.data["speed"] /= steps
        self.data["ego_linear_jerk"] /= steps
        self.data["ego_angular_jerk"] /= steps
        self.data["ego_num_violations"] /= steps
        self.data["social_num_violations"] /= steps
        self.data["max_speed_violation"] /= steps



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import os
import shutil
import unittest

import ray

from ultra.scenarios.generate_scenarios import build_scenarios
from ultra.train import train


class UltraPackageTest(unittest.TestCase):
    # Put generated files and folders in this directory.
    OUTPUT_DIRECTORY = "tests/ultra_package_test/"

    def test_simple_train_run(self):
        save_dir = "tests/task/eval_test/"
        log_dir = os.path.join(UltraPackageTest.OUTPUT_DIRECTORY, "tests/logs")

        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)

        save_dir = "tests/task/eval_test/eval"

        try:
            build_scenarios(
                task="task00",
                level_name="eval_test",
                stopwatcher_behavior=None,
                stopwatcher_route=None,
                root_path="tests/scenarios",
                save_dir=save_dir,
            )
        except:
            self.assertTrue(False)

        policy_class = "ultra.baselines.sac:sac-v0"

        ray.shutdown()
        try:
            ray.init(ignore_reinit_error=True)
            ray.wait(
                [
                    train.remote(
                        scenario_info=("00", "eval_test"),
                        policy_classes=[policy_class],
                        num_episodes=1,
                        max_episode_steps=2,
                        eval_info={
                            "eval_rate": 1000,
                            "eval_episodes": 2,
                        },
                        timestep_sec=0.1,
                        headless=True,
                        seed=2,
                        log_dir=log_dir,
                    )
                ]
            )
            ray.shutdown()
            self.assertTrue(True)
        except ray.exceptions.WorkerCrashedError as err:
            print(err)
            ray.shutdown()
            self.assertTrue(False)

        save_dir = "tests/task/eval_test/"

        if len(os.listdir(save_dir)) != 0:
            self.assertTrue(True)
            shutil.rmtree(save_dir)
        else:
            self.assertTrue(False)

        if os.path.exists(log_dir):
            self.assertTrue(True)
            shutil.rmtree(log_dir)
        else:
            self.assertTrue(False)

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(UltraPackageTest.OUTPUT_DIRECTORY):
            shutil.rmtree(UltraPackageTest.OUTPUT_DIRECTORY)
        if os.path.exists("tests/task/eval_test/"):
            shutil.rmtree("tests/task/eval_test/")



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import unittest

import gym
import numpy as np
import ray

from smarts.core.controllers import ActionSpaceType
from smarts.zoo.registry import make
from ultra.baselines.adapter import BaselineAdapter
from ultra.baselines.agent_spec import BaselineAgentSpec

AGENT_ID = "001"
seed = 2


class AdapterTest(unittest.TestCase):
    def test_observation_features(self):
        @ray.remote(max_calls=1, num_gpus=0, num_cpus=1)
        def run_experiment():
            agent, env = prepare_test_env_agent()
            observations = env.reset()
            env.close()
            return observations

        ray.init(ignore_reinit_error=True)
        observations = ray.get(run_experiment.remote())
        ray.shutdown()
        self.assertTrue("low_dim_states" in observations[AGENT_ID])
        self.assertTrue("social_vehicles" in observations[AGENT_ID])

        self.assertEqual(observations[AGENT_ID]["low_dim_states"].dtype, "float32")
        self.assertEqual(observations[AGENT_ID]["social_vehicles"].dtype, "float32")

    def test_rewards_adapter(self):
        @ray.remote(max_calls=1, num_gpus=0, num_cpus=1)
        def run_experiment():
            agent, env = prepare_test_env_agent()
            observations = env.reset()
            state = observations[AGENT_ID]
            action = agent.act(state, explore=True)
            observations, rewards, dones, infos = env.step({AGENT_ID: action})
            env.close()
            return rewards

        ray.init(ignore_reinit_error=True)
        rewards = ray.get(run_experiment.remote())
        ray.shutdown()
        self.assertIsInstance(rewards, dict)
        self.assertIsInstance(rewards[AGENT_ID], float)

    def test_rewards_returns(self):
        @ray.remote(max_calls=1, num_gpus=0, num_cpus=1)
        def run_experiment():
            agent, env = prepare_test_env_agent()
            observations = env.reset()
            state = observations[AGENT_ID]
            action = agent.act(state, explore=True)
            observations, rewards, dones, infos = env.step({AGENT_ID: action})
            env.close()
            return rewards

        ray.init(ignore_reinit_error=True)
        rewards = ray.get(run_experiment.remote())
        ray.shutdown()
        self.assertTrue(isinstance(rewards, dict))
        self.assertTrue(isinstance(rewards[AGENT_ID], float))


def prepare_test_env_agent(headless=True):
    timestep_sec = 0.1
    # [throttle, brake, steering]
    policy_class = "ultra.baselines.sac:sac-v0"
    spec = make(locator=policy_class)
    env = gym.make(
        "ultra.env:ultra-v0",
        agent_specs={AGENT_ID: spec},
        scenario_info=("00", "easy"),
        headless=headless,
        timestep_sec=timestep_sec,
        seed=seed,
    )
    agent = spec.build_agent()
    return agent, env



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# some parts of this implementation is inspired by https://github.com/openai/spinningup
import torch
import numpy as np
import torch.nn as nn
from sys import path
from ultra.baselines.sac.sac.network import SACNetwork
import torch.nn.functional as F
import pathlib, os, yaml, copy
from ultra.utils.common import compute_sum_aux_losses, to_3d_action, to_2d_action
from smarts.core.agent import Agent
from ultra.baselines.common.replay_buffer import ReplayBuffer
from ultra.baselines.common.social_vehicle_config import get_social_vehicle_configs
from ultra.baselines.common.yaml_loader import load_yaml
from ultra.baselines.common.baseline_state_preprocessor import BaselineStatePreprocessor


class SACPolicy(Agent):
    def __init__(
        self,
        policy_params=None,
        checkpoint_dir=None,
    ):
        # print("LOADING THE PARAMS", policy_params, checkpoint_dir)
        self.policy_params = policy_params
        self.gamma = float(policy_params["gamma"])
        self.critic_lr = float(policy_params["critic_lr"])
        self.actor_lr = float(policy_params["actor_lr"])
        self.critic_update_rate = int(policy_params["critic_update_rate"])
        self.policy_update_rate = int(policy_params["policy_update_rate"])
        self.warmup = int(policy_params["warmup"])
        self.seed = int(policy_params["seed"])
        self.batch_size = int(policy_params["batch_size"])
        self.hidden_units = int(policy_params["hidden_units"])
        self.tau = float(policy_params["tau"])
        self.initial_alpha = float(policy_params["initial_alpha"])
        self.logging_freq = int(policy_params["logging_freq"])
        self.action_size = int(policy_params["action_size"])
        self.prev_action = np.zeros(self.action_size)

        # state preprocessing
        self.social_policy_hidden_units = int(
            policy_params["social_vehicles"].get("social_policy_hidden_units", 0)
        )
        self.social_capacity = int(
            policy_params["social_vehicles"].get("social_capacity", 0)
        )
        self.observation_num_lookahead = int(
            policy_params.get("observation_num_lookahead", 0)
        )
        self.social_policy_init_std = int(
            policy_params["social_vehicles"].get("social_policy_init_std", 0)
        )
        self.num_social_features = int(
            policy_params["social_vehicles"].get("num_social_features", 0)
        )
        self.social_vehicle_config = get_social_vehicle_configs(
            **policy_params["social_vehicles"]
        )

        self.social_vehicle_encoder = self.social_vehicle_config["encoder"]
        self.state_description = BaselineStatePreprocessor.get_state_description(
            policy_params["social_vehicles"],
            policy_params["observation_num_lookahead"],
            self.action_size,
        )
        self.social_feature_encoder_class = self.social_vehicle_encoder[
            "social_feature_encoder_class"
        ]
        self.social_feature_encoder_params = self.social_vehicle_encoder[
            "social_feature_encoder_params"
        ]

        # others
        self.checkpoint_dir = checkpoint_dir
        self.device_name = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.device = torch.device(self.device_name)
        self.save_codes = (
            policy_params["save_codes"] if "save_codes" in policy_params else None
        )
        self.memory = ReplayBuffer(
            buffer_size=int(policy_params["replay_buffer"]["buffer_size"]),
            batch_size=int(policy_params["replay_buffer"]["batch_size"]),
            device_name=self.device_name,
        )
        self.current_iteration = 0
        self.steps = 0
        self.init_networks()
        if checkpoint_dir:
            self.load(checkpoint_dir)

    @property
    def state_size(self):
        # Adjusting state_size based on number of features (ego+social)
        size = sum(self.state_description["low_dim_states"].values())
        if self.social_feature_encoder_class:
            size += self.social_feature_encoder_class(
                **self.social_feature_encoder_params
            ).output_dim
        else:
            size += self.social_capacity * self.num_social_features
        # adding the previous action
        size += self.action_size
        return size

    def init_networks(self):
        self.sac_net = SACNetwork(
            action_size=self.action_size,
            state_size=self.state_size,
            hidden_units=self.hidden_units,
            seed=self.seed,
            initial_alpha=self.initial_alpha,
            social_feature_encoder_class=self.social_feature_encoder_class,
            social_feature_encoder_params=self.social_feature_encoder_params,
        ).to(self.device_name)

        self.actor_optimizer = torch.optim.Adam(
            self.sac_net.actor.parameters(), lr=self.actor_lr
        )

        self.critic_optimizer = torch.optim.Adam(
            self.sac_net.critic.parameters(), lr=self.critic_lr
        )

        self.log_alpha_optimizer = torch.optim.Adam(
            [self.sac_net.log_alpha], lr=self.critic_lr
        )

    def act(self, state, explore=True):
        state = copy.deepcopy(state)
        state["low_dim_states"] = np.float32(
            np.append(state["low_dim_states"], self.prev_action)
        )
        state["social_vehicles"] = (
            torch.from_numpy(state["social_vehicles"]).unsqueeze(0).to(self.device)
        )
        state["low_dim_states"] = (
            torch.from_numpy(state["low_dim_states"]).unsqueeze(0).to(self.device)
        )

        action, _, mean = self.sac_net.sample(state)

        if explore:  # training mode
            action = torch.squeeze(action, 0)
            action = action.detach().cpu().numpy()
        else:  # testing mode
            mean = torch.squeeze(mean, 0)
            action = mean.detach().cpu().numpy()
        return to_3d_action(action)

    def step(self, state, action, reward, next_state, done, info):
        # dont treat timeout as done equal to True
        max_steps_reached = info["logs"]["events"].reached_max_episode_steps
        if max_steps_reached:
            done = False
        action = to_2d_action(action)
        self.memory.add(
            state=state,
            action=action,
            reward=reward,
            next_state=next_state,
            done=float(done),
            social_capacity=self.social_capacity,
            observation_num_lookahead=self.observation_num_lookahead,
            social_vehicle_config=self.social_vehicle_config,
            prev_action=self.prev_action,
        )
        self.steps += 1
        output = {}
        if self.steps > max(self.warmup, self.batch_size):
            states, actions, rewards, next_states, dones, others = self.memory.sample(
                device=self.device_name
            )
            if self.steps % self.critic_update_rate == 0:
                critic_loss = self.update_critic(
                    states, actions, rewards, next_states, dones
                )
                output["loss/critic_loss"] = {
                    "type": "scalar",
                    "data": critic_loss.item(),
                    "freq": 2,
                }

            if self.steps % self.policy_update_rate == 0:
                actor_loss, temp_loss = self.update_actor_temp(
                    states, actions, rewards, next_states, dones
                )
                output["loss/actor_loss"] = {
                    "type": "scalar",
                    "data": actor_loss.item(),
                    "freq": self.logging_freq,
                }
                output["loss/temp_loss"] = {
                    "type": "scalar",
                    "data": temp_loss.item(),
                    "freq": self.logging_freq,
                }
                output["others/alpha"] = {
                    "type": "scalar",
                    "data": self.sac_net.alpha.item(),
                    "freq": self.logging_freq,
                }
                self.current_iteration += 1
            self.target_soft_update(self.sac_net.critic, self.sac_net.target, self.tau)
        self.prev_action = action if not done else np.zeros(self.action_size)
        return output

    def update_critic(self, states, actions, rewards, next_states, dones):

        q1_current, q2_current, aux_losses = self.sac_net.critic(
            states, actions, training=True
        )
        with torch.no_grad():
            next_actions, log_probs, _ = self.sac_net.sample(next_states)
            q1_next, q2_next = self.sac_net.target(next_states, next_actions)
            v_next = (
                torch.min(q1_next, q2_next) - self.sac_net.alpha.detach() * log_probs
            )
            q_target = (rewards + ((1 - dones) * self.gamma * v_next)).detach()

        critic_loss = F.mse_loss(q1_current, q_target) + F.mse_loss(
            q2_current, q_target
        )

        aux_losses = compute_sum_aux_losses(aux_losses)
        overall_loss = critic_loss + aux_losses
        self.critic_optimizer.zero_grad()
        overall_loss.backward()
        self.critic_optimizer.step()

        return critic_loss

    def update_actor_temp(self, states, actions, rewards, next_states, dones):

        for p in self.sac_net.target.parameters():
            p.requires_grad = False
        for p in self.sac_net.critic.parameters():
            p.requires_grad = False

        # update actor:
        actions, log_probs, aux_losses = self.sac_net.sample(states, training=True)
        q1, q2 = self.sac_net.critic(states, actions)
        q_old = torch.min(q1, q2)
        actor_loss = (self.sac_net.alpha.detach() * log_probs - q_old).mean()
        aux_losses = compute_sum_aux_losses(aux_losses)
        overall_loss = actor_loss + aux_losses
        self.actor_optimizer.zero_grad()
        overall_loss.backward()
        self.actor_optimizer.step()

        # update temp:
        temp_loss = (
            self.sac_net.log_alpha.exp()
            * (-log_probs.detach().mean() + self.action_size).detach()
        )
        self.log_alpha_optimizer.zero_grad()
        temp_loss.backward()
        self.log_alpha_optimizer.step()
        self.sac_net.alpha.data = self.sac_net.log_alpha.exp().detach()

        for p in self.sac_net.target.parameters():
            p.requires_grad = True
        for p in self.sac_net.critic.parameters():
            p.requires_grad = True

        return actor_loss, temp_loss

    def target_soft_update(self, critic, target_critic, tau):
        with torch.no_grad():
            for critic_param, target_critic_param in zip(
                critic.parameters(), target_critic.parameters()
            ):
                target_critic_param.data = (
                    tau * critic_param.data + (1 - tau) * target_critic_param.data
                )

    def load(self, model_dir):
        model_dir = pathlib.Path(model_dir)
        map_location = None
        if self.device and self.device.type == "cpu":
            map_location = "cpu"
        self.sac_net.actor.load_state_dict(
            torch.load(model_dir / "actor.pth", map_location=map_location)
        )
        self.sac_net.target.load_state_dict(
            torch.load(model_dir / "target.pth", map_location=map_location)
        )
        self.sac_net.critic.load_state_dict(
            torch.load(model_dir / "critic.pth", map_location=map_location)
        )
        print("<<<<<<< MODEL LOADED >>>>>>>>>", model_dir)

    def save(self, model_dir):
        model_dir = pathlib.Path(model_dir)
        # with open(model_dir / "params.yaml", "w") as file:
        #     yaml.dump(policy_params, file)

        torch.save(self.sac_net.actor.state_dict(), model_dir / "actor.pth")
        torch.save(self.sac_net.target.state_dict(), model_dir / "target.pth")
        torch.save(self.sac_net.critic.state_dict(), model_dir / "critic.pth")
        print("<<<<<<< MODEL SAVED >>>>>>>>>", model_dir)

    def reset(self):
        pass



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import glob
import subprocess
import unittest

import ray

from ultra.scenarios.analysis.base_analysis import BaseAnalysis
from ultra.scenarios.analysis.scenario_analysis import ScenarioAnalysis
from ultra.scenarios.analysis.sumo_experiment import (
    edge_lane_data_function,
    sumo_rerouting_routine,
    vehicle_data_function,
)
from ultra.scenarios.generate_scenarios import *


class AnalysisTest(unittest.TestCase):
    # Put generated files and folders in this directory.
    OUTPUT_DIRECTORY = "tests/analysis_test/"

    def test_interface_analyze(self):
        try:
            save_dir = os.path.join(AnalysisTest.OUTPUT_DIRECTORY, "scenarios/")
            output = os.path.join(AnalysisTest.OUTPUT_DIRECTORY, "output/")
            if os.path.exists(save_dir):
                shutil.rmtree(save_dir)
            if os.path.exists(output):
                shutil.rmtree(save_dir)
            build_scenarios(
                task="task00",
                level_name="easy",
                stopwatcher_behavior="aggressive",
                stopwatcher_route="south-west",
                root_path="tests/scenarios",
                save_dir=save_dir,
            )

            if not os.path.exists(output):
                os.makedirs(output)

            os.system(
                f"python ultra/scenarios/interface.py analyze --scenarios {save_dir} --max-steps 600 --end-by-stopwatcher --output {output}"
            )
            for dirpath, dirnames, files in os.walk(save_dir):
                if "traffic" in dirpath:
                    self.assertTrue("all.rou.xml" in files)

            self.assertTrue(os.path.exists("tests/analysis_test/output/analysis.pkl"))
        except Exception as err:
            print(err)
            self.assertTrue(False)

        if os.path.exists(save_dir):
            self.assertTrue(True)
            shutil.rmtree(save_dir)

        if os.path.exists(output):
            self.assertTrue(True)
            shutil.rmtree(output)

    def test_analyze_scenario(self):
        save_dir = os.path.join(AnalysisTest.OUTPUT_DIRECTORY, "scenarios/")
        output = os.path.join(AnalysisTest.OUTPUT_DIRECTORY, "output/")
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        if os.path.exists(output):
            shutil.rmtree(output)

        build_scenarios(
            task="task00",
            level_name="easy",
            stopwatcher_behavior="aggressive",
            stopwatcher_route="south-west",
            root_path="tests/scenarios",
            save_dir=save_dir,
        )
        scenarios = glob.glob(f"{save_dir}")
        try:
            ray.init(ignore_reinit_error=True)
            analyzer = ScenarioAnalysis.remote()
            ray.wait(
                [
                    analyzer.run.remote(
                        end_by_stopwatcher=True,
                        scenarios=scenarios,
                        ego=None,
                        max_episode_steps=800,
                        policy=None,
                        video_rate=100,
                        timestep_sec=0.1,
                        custom_traci_functions=[
                            edge_lane_data_function,
                            vehicle_data_function,
                        ],
                    )
                ]
            )
            if not os.path.exists(output):
                os.makedirs(output)
            ray.wait([analyzer.save_data.remote(save_dir=output)])
            self.assertTrue(os.path.exists(f"{output}analysis.pkl"))

            for dirpath, dirnames, files in os.walk(save_dir):
                if "traffic" in dirpath:
                    self.assertTrue("all.rou.xml" in files)

            self.assertTrue(os.path.exists("tests/analysis_test/output/analysis.pkl"))
        except ray.exceptions.WorkerCrashedError as err:
            print(err)
            self.assertTrue(False)

        if os.path.exists(save_dir):
            self.assertTrue(True)
            shutil.rmtree(save_dir)

        if os.path.exists(output):
            self.assertTrue(True)
            shutil.rmtree(output)

    def test_save_histogram(self):
        try:
            figure_name = os.path.join(AnalysisTest.OUTPUT_DIRECTORY, "histogram")
            if not os.path.exists(figure_name):
                os.makedirs(figure_name)
            analyzer = BaseAnalysis()
            data = [1, 2, 3, 4]

            analyzer.save_histogram(data=data, figure_name=figure_name, title="test")
            if not os.path.exists(figure_name):
                os.makedirs(figure_name)

        except Exception as err:
            print(err)
            self.assertTrue(False)

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(AnalysisTest.OUTPUT_DIRECTORY):
            shutil.rmtree(AnalysisTest.OUTPUT_DIRECTORY)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from collections import defaultdict
from glob import glob
from os import path
from pathlib import Path

from setuptools import find_packages, setup

this_dir = path.abspath(path.dirname(__file__))
with open(path.join(this_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


""" Modified setup.py to include option for changing SMARTS version or, by default,
the latest stable version SMARTS will used """
setup(
    name="ultra-rl",
    description="Unprotected Left Turn for Robust Agents",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="0.2.0",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    zip_safe=True,
    python_requires=">=3.7",
    install_requires=[
        "smarts[train,test]==0.4.14",
        "setuptools>=41.0.0,!=50.0",
        "dill",
        "black==20.8b1",
        "ray[rllib]==1.0.1.post1",
    ],
)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# Do not make any change to this file when merging. Just use my version.
from collections import deque, namedtuple
import numpy as np
import random, copy
import torch
from ultra.utils.common import normalize_im
from collections.abc import Iterable

from torch.utils.data import Dataset, Sampler, DataLoader

Transition = namedtuple(
    "Transition",
    field_names=["state", "action", "reward", "next_state", "done", "others"],
    # others may contain importance sampling ratio, GVF rewards,... etc
    defaults=(None,) * 6,
)


class RandomRLSampler(Sampler):
    def __init__(self, data_source, batch_size):
        self.datasource = data_source
        self.batch_size = batch_size

    def __len__(self):
        return len(self.datasource)

    def __iter__(self):
        n = len(self.datasource)
        return iter(torch.randperm(n).tolist()[0 : self.batch_size])


class ReplayBufferDataset(Dataset):
    cpu = torch.device("cpu")

    def __init__(self, buffer_size, device):
        self.buffer_size = buffer_size
        self.memory = deque(maxlen=self.buffer_size)
        self.device = device

    def add(
        self,
        state,
        action,
        reward,
        next_state,
        done,
        social_capacity,
        observation_num_lookahead,
        social_vehicle_config,
        prev_action,
        others=None,
    ):
        if others is None:
            others = {}
        # dereference the states
        state = copy.deepcopy(state)
        next_state = copy.deepcopy(next_state)
        state["low_dim_states"] = np.float32(
            np.append(state["low_dim_states"], prev_action)
        )
        state["low_dim_states"] = torch.from_numpy(state["low_dim_states"]).to(
            self.device
        )
        state["social_vehicles"] = torch.from_numpy(state["social_vehicles"]).to(
            self.device
        )

        next_state["low_dim_states"] = np.float32(
            np.append(next_state["low_dim_states"], action)
        )
        next_state["social_vehicles"] = torch.from_numpy(
            next_state["social_vehicles"]
        ).to(self.device)
        next_state["low_dim_states"] = torch.from_numpy(
            next_state["low_dim_states"]
        ).to(self.device)

        action = np.asarray([action]) if not isinstance(action, Iterable) else action
        action = torch.from_numpy(action).float()
        reward = torch.from_numpy(np.asarray([reward])).float()
        done = torch.from_numpy(np.asarray([done])).float()
        new_experience = Transition(state, action, reward, next_state, done, others)
        self.memory.append(new_experience)

    def __len__(self):
        return len(self.memory)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        state, action, reward, next_state, done, others = tuple(self.memory[idx])
        return state, action, reward, next_state, done, others


class ReplayBuffer:
    def __init__(
        self,
        buffer_size,
        batch_size,
        device_name,
        pin_memory=False,
        num_workers=0,
    ):
        self.replay_buffer_dataset = ReplayBufferDataset(buffer_size, device=None)
        self.sampler = RandomRLSampler(self.replay_buffer_dataset, batch_size)
        self.data_loader = DataLoader(
            self.replay_buffer_dataset,
            sampler=self.sampler,
            pin_memory=pin_memory,
            num_workers=num_workers,
        )
        self.storage_device = torch.device(device_name)

    def add(self, *args, **kwargs):
        self.replay_buffer_dataset.add(*args, **kwargs)

    def __len__(self):
        return len(self.replay_buffer_dataset)

    def __getitem__(self, idx):
        return self.replay_buffer_dataset[idx]

    def make_state_from_dict(self, states, device):
        # image_keys = states[0]["images"].keys()
        # images = {}
        # for k in image_keys:
        #     _images = torch.cat([e[k] for e in states], dim=0).float().to(device)
        #     _images = normalize_im(_images)
        #     images[k] = _images
        low_dim_states = (
            torch.cat([e["low_dim_states"] for e in states], dim=0).float().to(device)
        )
        if "social_vehicles" in states[0]:
            social_vehicles = [
                e["social_vehicles"][0].float().to(device) for e in states
            ]
        else:
            social_vehicles = False
        out = {
            # "images": images,
            "low_dim_states": low_dim_states,
            "social_vehicles": social_vehicles,
        }
        return out

    def sample(self, device=None):
        device = device if device else self.storage_device
        batch = list(iter(self.data_loader))
        states, actions, rewards, next_states, dones, others = zip(*batch)
        states = self.make_state_from_dict(states, device)
        next_states = self.make_state_from_dict(next_states, device)
        actions = torch.cat(actions, dim=0).float().to(device)
        rewards = torch.cat(rewards, dim=0).float().to(device)
        dones = torch.cat(dones, dim=0).float().to(device)
        return states, actions, rewards, next_states, dones, others
